@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2003-2018. All Rights Reserved.
+ * Copyright Ericsson AB 2003-2020. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -58,6 +58,7 @@ static void dump_externally(fmtfn_t to, void *to_arg, Eterm term);
 static void mark_literal(Eterm* ptr);
 static void init_literal_areas(void);
 static void dump_literals(fmtfn_t to, void *to_arg);
+static void dump_persistent_terms(fmtfn_t to, void *to_arg);
 static void dump_module_literals(fmtfn_t to, void *to_arg,
                                  ErtsLiteralArea* lit_area);
 
@@ -74,6 +75,7 @@ erts_deep_process_dump(fmtfn_t to, void *to_arg)
 
     all_binaries = NULL;
     init_literal_areas();
+    erts_init_persistent_dumping();
 
     for (i = 0; i < max; i++) {
 	Process *p = erts_pix2proc(i);
@@ -93,20 +95,23 @@ erts_deep_process_dump(fmtfn_t to, void *to_arg)
        }
     }
 
+    dump_persistent_terms(to, to_arg);
     dump_literals(to, to_arg);
     dump_binaries(to, to_arg, all_binaries);
 }
 
-static void
-monitor_size(ErtsMonitor *mon, void *vsize)
+static int
+monitor_size(ErtsMonitor *mon, void *vsize, Sint reds)
 {
     *((Uint *) vsize) += erts_monitor_size(mon);
+    return 1;
 }
 
-static void
-link_size(ErtsMonitor *lnk, void *vsize)
+static int
+link_size(ErtsMonitor *lnk, void *vsize, Sint reds)
 {
     *((Uint *) vsize) += erts_link_size(lnk);
+    return 1;
 }
 
 Uint erts_process_memory(Process *p, int include_sigs_in_transit)
@@ -116,12 +121,14 @@ Uint erts_process_memory(Process *p, int include_sigs_in_transit)
 
     size += sizeof(Process);
 
-    erts_link_tree_foreach(ERTS_P_LINKS(p),
-                           link_size, (void *) &size);
-    erts_monitor_tree_foreach(ERTS_P_MONITORS(p),
-                              monitor_size, (void *) &size);
-    erts_monitor_list_foreach(ERTS_P_LT_MONITORS(p),
-                              monitor_size, (void *) &size);
+    if ((erts_atomic32_read_nob(&p->state) & ERTS_PSFLG_EXITING) == 0) {
+        erts_link_tree_foreach(ERTS_P_LINKS(p),
+                               link_size, (void *) &size);
+        erts_monitor_tree_foreach(ERTS_P_MONITORS(p),
+                                  monitor_size, (void *) &size);
+        erts_monitor_list_foreach(ERTS_P_LT_MONITORS(p),
+                                  monitor_size, (void *) &size);
+    }
     size += (p->heap_sz + p->mbuf_sz) * sizeof(Eterm);
     if (p->abandoned_heap)
         size += (p->hend - p->heap) * sizeof(Eterm);
@@ -186,11 +193,11 @@ static ERTS_INLINE void
 dump_msg(fmtfn_t to, void *to_arg, ErtsMessage *mp)
 {
     if (ERTS_SIG_IS_MSG((ErtsSignal *) mp)) {
-        Eterm mesg = ERL_MESSAGE_TERM(mp);
-        if (is_value(mesg))
-            dump_element(to, to_arg, mesg);
+        Eterm mesg;
+        if (ERTS_SIG_IS_INTERNAL_MSG(mp))
+            dump_element(to, to_arg, ERL_MESSAGE_TERM(mp));
         else
-            dump_dist_ext(to, to_arg, mp->data.dist_ext);
+            dump_dist_ext(to, to_arg, erts_get_dist_ext(mp->data.heap_frag));
         mesg = ERL_MESSAGE_TOKEN(mp);
         erts_print(to, to_arg, ":");
         dump_element(to, to_arg, mesg);
@@ -262,6 +269,7 @@ dump_dist_ext(fmtfn_t to, void *to_arg, ErtsDistExternal *edep)
     else {
 	byte *e;
 	size_t sz;
+        int i;
 
 	if (!(edep->flags & ERTS_DIST_EXT_ATOM_TRANS_TAB))
 	    erts_print(to, to_arg, "D0:");
@@ -271,8 +279,8 @@ dump_dist_ext(fmtfn_t to, void *to_arg, ErtsDistExternal *edep)
 	    for (i = 0; i < edep->attab.size; i++)
 		dump_element(to, to_arg, edep->attab.atom[i]);
 	}
-	sz = edep->ext_endp - edep->extp;
-	e = edep->extp;
+	sz = edep->data->ext_endp - edep->data->extp;
+	e = edep->data->extp;
 	if (edep->flags & ERTS_DIST_EXT_DFLAG_HDR) {
 	    ASSERT(*e != VERSION_MAGIC);
 	    sz++;
@@ -283,15 +291,19 @@ dump_dist_ext(fmtfn_t to, void *to_arg, ErtsDistExternal *edep)
 	erts_print(to, to_arg, "E%X:", sz);
         if (edep->flags & ERTS_DIST_EXT_DFLAG_HDR) {
             byte sbuf[3];
-            int i = 0;
+
+            i = 0;
 
             sbuf[i++] = VERSION_MAGIC;
-            while (i < sizeof(sbuf) && e < edep->ext_endp) {
+            while (i < sizeof(sbuf) && e < edep->data->ext_endp) {
                 sbuf[i++] = *e++;
             }
             erts_print_base64(to, to_arg, sbuf, i);
         }
-        erts_print_base64(to, to_arg, e, edep->ext_endp - e);
+        erts_print_base64(to, to_arg, e, edep->data->ext_endp - e);
+        for (i = 1; i < edep->data->frag_id; i++)
+            erts_print_base64(to, to_arg, edep->data[i].extp,
+                              edep->data[i].ext_endp - edep->data[i].extp);
     }
 }
 
@@ -775,6 +787,9 @@ init_literal_areas(void)
     qsort(lit_areas, num_lit_areas, sizeof(ErtsLiteralArea *),
           compare_areas);
 
+    qsort(erts_persistent_areas, erts_num_persistent_areas,
+          sizeof(ErtsLiteralArea *), compare_areas);
+
     erts_runlock_old_code(code_ix);
 }
 
@@ -796,6 +811,13 @@ static void mark_literal(Eterm* ptr)
 
     ap = bsearch(ptr, lit_areas, num_lit_areas, sizeof(ErtsLiteralArea*),
                  search_areas);
+    if (ap == 0) {
+        ap = bsearch(ptr, erts_persistent_areas,
+                     erts_num_persistent_areas,
+                     sizeof(ErtsLiteralArea*),
+                     search_areas);
+    }
+
 
     /*
      * If the literal was created by native code, this search will not
@@ -807,12 +829,12 @@ static void mark_literal(Eterm* ptr)
     }
 }
 
-
 static void
 dump_literals(fmtfn_t to, void *to_arg)
 {
     ErtsCodeIndex code_ix;
     int i;
+    Uint idx;
 
     code_ix = erts_active_code_ix();
     erts_rlock_old_code(code_ix);
@@ -825,6 +847,28 @@ dump_literals(fmtfn_t to, void *to_arg)
     }
 
     erts_runlock_old_code(code_ix);
+
+    for (idx = 0; idx < erts_num_persistent_areas; idx++) {
+        dump_module_literals(to, to_arg, erts_persistent_areas[idx]);
+    }
+}
+
+static void
+dump_persistent_terms(fmtfn_t to, void *to_arg)
+{
+    Uint idx;
+
+    erts_print(to, to_arg, "=persistent_terms\n");
+
+    for (idx = 0; idx < erts_num_persistent_areas; idx++) {
+        ErtsLiteralArea* ap = erts_persistent_areas[idx];
+        Eterm tuple = make_tuple(ap->start);
+        Eterm* tup_val = tuple_val(tuple);
+
+	dump_element(to, to_arg, tup_val[1]);
+        erts_putc(to, to_arg, '|');
+	dump_element_nl(to, to_arg, tup_val[2]);
+    }
 }
 
 static void
@@ -963,12 +1007,17 @@ dump_module_literals(fmtfn_t to, void *to_arg, ErtsLiteralArea* lit_area)
                     }
                     erts_putc(to, to_arg, '\n');
                 }
-            } else if (is_export_header(w)) {
+            } else {
+                /* Dump everything else in the external format */
                 dump_externally(to, to_arg, term);
                 erts_putc(to, to_arg, '\n');
             }
             size = 1 + header_arity(w);
             switch (w & _HEADER_SUBTAG_MASK) {
+            case FUN_SUBTAG:
+                ASSERT(((ErlFunThing*)(htop))->num_free == 0);
+                size += 1;
+                break;
             case MAP_SUBTAG:
                 if (is_flatmap_header(w)) {
                     size += 1 + flatmap_get_size(htop);
