@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 2006-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2006-2020. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -32,13 +32,14 @@
 
 -include_lib("common_test/include/ct.hrl").
 
--export([all/0, suite/0]).
+-export([all/0, suite/0, init_per_testcase/2, end_per_testcase/2]).
 
 -export([schedulers_alive/1, node_container_refc_check/1,
 	 long_timers/1, pollset_size/1,
 	 check_io_debug/1, get_check_io_info/0,
          lc_graph/1,
-         leaked_processes/1]).
+         leaked_processes/1,
+         literal_area_collector/1]).
 
 suite() ->
     [{ct_hooks,[ts_install_cth]},
@@ -50,7 +51,22 @@ all() ->
      lc_graph,
      %% Make sure that the leaked_processes/1 is always
      %% run last.
-     leaked_processes].
+     leaked_processes,
+     literal_area_collector].
+
+init_per_testcase(schedulers_alive, Config) ->
+    case erlang:system_info(schedulers) of
+        1 ->
+            {skip, "Needs more schedulers to run"};
+        _ ->
+            Config
+    end;
+init_per_testcase(_, Config) ->
+    Config.
+
+
+end_per_testcase(_Name, Config) ->
+    Config.
 
 %%%
 %%% The test cases -------------------------------------------------------------
@@ -251,7 +267,7 @@ pollset_size(Config) when is_list(Config) ->
 	  end.
 
 check_io_debug(Config) when is_list(Config) ->
-    case lists:keysearch(name, 1, erlang:system_info(check_io)) of
+    case lists:keysearch(name, 1, hd(erlang:system_info(check_io))) of
 	      {value, {name, erts_poll}} -> check_io_debug_test();
 	      _ -> {skipped, "Not implemented in this emulator"}
 	  end.
@@ -323,10 +339,14 @@ leaked_processes(Config) when is_list(Config) ->
                                           [length(Leaked)])),
     {comment, Comment}.
 
+literal_area_collector(Config) when is_list(Config) ->
+    literal_area_collector_test:check_idle(10000).
+
 %%
 %% Internal functions...
 %%
 
+
 display_check_io(ChkIo) ->
     catch erlang:display('--- CHECK IO INFO ---'),
     catch erlang:display(ChkIo),
