@@ -1,7 +1,7 @@
 -module(diameter_dict_parser).
 -export([parse/1, parse_and_scan/1, format_error/1]).
 
--file("/net/isildur/ldisk/daily_build/22_prebuild_opu_o.2019-09-17_11/otp_src_22/bootstrap/lib/parsetools/include/yeccpre.hrl", 0).
+-file("/net/isildur/ldisk/daily_build/23_prebuild_master-opu_o.2020-05-12_22/otp_src_23/bootstrap/lib/parsetools/include/yeccpre.hrl", 0).
 %%
 %% %CopyrightBegin%
 %%
