@@ -21,8 +21,10 @@
 -module(crashdump_helper).
 -export([n1_proc/2,remote_proc/2,
          dump_maps/0,create_maps/0,
-         create_binaries/0,create_sub_binaries/1]).
--compile(r18).
+         create_binaries/0,create_sub_binaries/1,
+         dump_persistent_terms/0,
+         create_persistent_terms/0]).
+-compile(r20).
 -include_lib("common_test/include/ct.hrl").
 
 n1_proc(N2,Creator) ->
@@ -46,7 +48,7 @@ n1_proc(Creator,_N2,Pid2,Port2,_L) ->
     Ref = make_ref(),
     Pid = self(),
     Bin = list_to_binary(lists:seq(1, 255)),
-    <<_:2,SubBin:17/binary,_/bits>> = Bin,
+    <<_:2,SubBin:65/binary,_/bits>> = Bin,
 
     register(named_port,Port),
 
@@ -62,6 +64,7 @@ n1_proc(Creator,_N2,Pid2,Port2,_L) ->
     put(ref,Ref),
     put(pid,Pid),
     put(bin,Bin),
+    put(proc_bins,create_proc_bins()),
     put(bins,create_binaries()),
     put(sub_bin,SubBin),
     put(sub_bins,create_sub_binaries(get(bins))),
@@ -99,7 +102,7 @@ remote_proc(P1,Creator) ->
 	  end).
 
 create_binaries() ->
-    Sizes = lists:seq(60, 70) ++ lists:seq(120, 140),
+    Sizes = lists:seq(100, 120) ++ lists:seq(200, 240),
     [begin
          <<H:16/unit:8>> = erlang:md5(<<Size:32>>),
          Data = ((H bsl (8*150)) div (H+7919)),
@@ -110,13 +113,30 @@ create_sub_binaries(Bins) ->
     [create_sub_binary(Bin, Start, LenSub) ||
         Bin <- Bins,
         Start <- [0,1,2,3,4,5,10,22],
-        LenSub <- [0,1,2,3,4,6,9]].
+        LenSub <- [0,1,2,3,4,6,9,65]].
 
 create_sub_binary(Bin, Start, LenSub) ->
     Len = byte_size(Bin) - LenSub - Start,
     <<_:Start/bytes,Sub:Len/bytes,_/bytes>> = Bin,
     Sub.
 
+create_proc_bins() ->
+    Parent = self(),
+    Pid =
+        spawn(
+          fun() ->
+                  %% Just reverse the list here, so this binary is not
+                  %% confused with the one created in n1_proc/5 above,
+                  %% which is used for testing truncation (see
+                  %% crashdump_viewer_SUITE:truncate_dump_binary/1)
+                  Bin = list_to_binary(lists:reverse(lists:seq(1, 255))),
+                  <<A:65/bytes,B:65/bytes,C/bytes>> = Bin,
+                  Parent ! {self(),{A,B,C}}
+          end),
+    receive
+        {Pid,ProcBins} -> ProcBins
+    end.
+
 %%%
 %%% Test dumping of maps. Dumping of maps only from OTP 20.2.
 %%%
@@ -132,6 +152,7 @@ dump_maps() ->
     Pid = spawn_link(F),
     receive
         {Pid,done} ->
+            unlink(Pid),
             {ok,Pid}
     end.
 
@@ -142,4 +163,47 @@ create_maps() ->
     Map3 = lists:foldl(fun(I, A) ->
                                A#{I=>I*I}
                        end, Map2, lists:seq(-10, 0)),
-    #{a=>Map0,b=>Map1,c=>Map2,d=>Map3,e=>#{}}.
+    #{a=>Map0,b=>Map1,c=>Map2,d=>Map3,e=>#{},literal=>literal_map()}.
+
+literal_map() ->
+    %% A literal map such as the one below will produce a heap dump
+    %% like this:
+    %%
+    %%   Address1:t4:H<Address3>,H<Address4>,H<Address5>,H<Address6>
+    %%   Address2:Mf4:H<Adress1>:I1,I2,I3,I4
+    %%   Address3: ...  % "one"
+    %%   Address4: ...  % "two"
+    %%   Address5: ...  % "three"
+    %%   Address6: ...  % "four"
+    %%
+    %% The map cannot be reconstructed in a single sequential pass.
+    %%
+    %% To reconstruct the map, first the string keys "one"
+    %% through "four" must be reconstructed, then the tuple at
+    %% Adress1, then the map at Address2.
+
+    #{"one"=>1,"two"=>2,"three"=>3,"four"=>4}.
+
+%%%
+%%% Test dumping of persistent terms (from OTP 21.2).
+%%%
+
+dump_persistent_terms() ->
+    Parent = self(),
+    F = fun() ->
+                register(aaaaaaaa_persistent_terms, self()),
+                put(pts, create_persistent_terms()),
+                Parent ! {self(),done},
+                receive _ -> ok end
+        end,
+    Pid = spawn_link(F),
+    receive
+        {Pid,done} ->
+            unlink(Pid),
+            {ok,Pid}
+    end.
+
+create_persistent_terms() ->
+    persistent_term:put({?MODULE,first}, {pid,42.0}),
+    persistent_term:put({?MODULE,second}, [1,2,3]),
+    {persistent_term:get({?MODULE,first}),persistent_term:get({?MODULE,second})}.
