@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2004-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2004-2020. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -28,6 +28,7 @@
 -include_lib("public_key/include/public_key.hrl").
 -include_lib("common_test/include/ct.hrl").
 -include_lib("ssh/src/ssh_transport.hrl").
+-include_lib("kernel/include/file.hrl").
 -include("ssh_test_lib.hrl").
 
 %%%----------------------------------------------------------------
@@ -163,7 +164,7 @@ std_simple_exec(Host, Port, Config, Opts) ->
     ct:log("~p:~p exec ~p",[?MODULE,?LINE,ExecResult]),
     case ExecResult of
 	success ->
-	    Expected = {ssh_cm, ConnectionRef, {data,ChannelId,0,<<"42\n">>}},
+	    Expected = {ssh_cm, ConnectionRef, {data,ChannelId,0,<<"42">>}},
 	    case receive_exec_result(Expected) of
 		expected ->
 		    ok;
@@ -290,44 +291,57 @@ rcv_lingering(Timeout) ->
     end.
 
 
-receive_exec_result(Msg) ->
-    ct:log("Expect data! ~p", [Msg]),
+receive_exec_result([]) ->
+    expected;
+receive_exec_result(Msgs) when is_list(Msgs) ->
+    ct:log("~p:~p Expect data! ~p", [?MODULE,?FUNCTION_NAME,Msgs]),
     receive
-	{ssh_cm,_,{data,_,1, Data}} ->
-	    ct:log("StdErr: ~p~n", [Data]),
-	    receive_exec_result(Msg);
-	Msg ->
-	    ct:log("1: Collected data ~p", [Msg]),
-	    expected;
-	Other ->
-	    ct:log("Other ~p", [Other]),
-	    {unexpected_msg, Other}
+        Msg ->
+            case lists:member(Msg, Msgs)
+                orelse lists:member({optional,Msg}, Msgs)
+            of
+                true ->
+                    ct:log("~p:~p Collected data ~p", [?MODULE,?FUNCTION_NAME,Msg]),
+                    receive_exec_result(Msgs--[Msg,{optional,Msg}]);
+                false ->
+                    case Msg of
+                        {ssh_cm,_,{data,_,1, Data}} ->
+                            ct:log("~p:~p StdErr: ~p~n", [?MODULE,?FUNCTION_NAME,Data]),
+                            receive_exec_result(Msgs);
+                        Other ->
+                            ct:log("~p:~p Other ~p", [?MODULE,?FUNCTION_NAME,Other]),
+                            {unexpected_msg, Other}
+                    end
+            end
     after 
-	30000 -> ct:fail("timeout ~p:~p",[?MODULE,?LINE])
-    end.
+	30000 ->
+            case lists:all(fun(M) ->
+                                   is_tuple(M) andalso (element(1,M) == optional)
+                           end, Msgs)
+            of
+                false ->
+                    ct:fail("timeout ~p:~p",[?MODULE,?FUNCTION_NAME]);
+                true ->
+                    ct:log("~p:~p Only optional messages expected!~n ~p", [?MODULE,?FUNCTION_NAME,Msgs]),
+                    expected
+            end
+    end;
+receive_exec_result(Msg) ->
+    receive_exec_result([Msg]).
 
 
+receive_exec_result_or_fail(Msg) ->
+    case receive_exec_result(Msg) of
+        expected -> expected;
+        Other -> ct:fail(Other)
+    end.
+
 receive_exec_end(ConnectionRef, ChannelId) ->
-    Eof = {ssh_cm, ConnectionRef, {eof, ChannelId}},
-    ExitStatus = {ssh_cm, ConnectionRef, {exit_status, ChannelId, 0}},
-    Closed =  {ssh_cm, ConnectionRef,{closed, ChannelId}},
-    case receive_exec_result(ExitStatus) of
-	{unexpected_msg, Eof} -> %% Open ssh seems to not allways send these messages
-	    %% in the same order!
-	    ct:log("2: Collected data ~p", [Eof]),
-	    case receive_exec_result(ExitStatus) of
-		expected ->
-		    expected = receive_exec_result(Closed);
-		{unexpected_msg, Closed} ->
-		    ct:log("3: Collected data ~p", [Closed])
-	    end;
-	expected ->
-	    ct:log("4: Collected data ~p", [ExitStatus]),
-	    expected = receive_exec_result(Eof),
-	    expected = receive_exec_result(Closed);
-	Other ->
-	    ct:fail({unexpected_msg, Other})
-    end.
+    receive_exec_result(
+      [{ssh_cm, ConnectionRef, {eof, ChannelId}},
+       {optional, {ssh_cm, ConnectionRef, {exit_status, ChannelId, 0}}},
+       {ssh_cm, ConnectionRef, {closed, ChannelId}}
+      ]).
 
 receive_exec_result(Data, ConnectionRef, ChannelId) ->
     Eof = {ssh_cm, ConnectionRef, {eof, ChannelId}},
@@ -337,27 +351,6 @@ receive_exec_result(Data, ConnectionRef, ChannelId) ->
     expected = receive_exec_result(Closed).
 
 
-inet_port()->
-    {ok, Socket} = gen_tcp:listen(0, [{reuseaddr, true}]),
-    {ok, Port} = inet:port(Socket),
-    gen_tcp:close(Socket),
-    Port.
-
-setup_ssh_auth_keys(RSAFile, DSAFile, Dir) ->
-    Entries = ssh_file_entry(RSAFile) ++ ssh_file_entry(DSAFile),
-    AuthKeys = public_key:ssh_encode(Entries , auth_keys),
-    AuthKeysFile = filename:join(Dir, "authorized_keys"),
-    file:write_file(AuthKeysFile, AuthKeys).
-
-ssh_file_entry(PubFile) ->
-    case file:read_file(PubFile) of
-	{ok, Ssh} ->
-	    [{Key, _}] = public_key:ssh_decode(Ssh, public_key), 
-	    [{Key, [{comment, "Test"}]}];
-	_ ->
-	    []
-    end. 
-	    
 failfun(_User, {authmethod,none}) ->
     ok;
 failfun(User, Reason) ->
@@ -367,210 +360,38 @@ hostname() ->
     {ok,Host} = inet:gethostname(),
     Host.
 
-known_hosts(BR) ->
-    KnownHosts = ssh_file:file_name(user, "known_hosts", []),
-    B = KnownHosts ++ "xxx",
-    case BR of
-	backup ->
-	    file:rename(KnownHosts, B);
-	restore ->
-	    file:delete(KnownHosts),
-	    file:rename(B, KnownHosts)
-    end.
-
-setup_dsa(DataDir, UserDir) ->
-    file:copy(filename:join(DataDir, "id_dsa"), filename:join(UserDir, "id_dsa")),
-    System = filename:join(UserDir, "system"),
-    file:make_dir(System),
-    file:copy(filename:join(DataDir, "ssh_host_dsa_key"), filename:join(System, "ssh_host_dsa_key")),
-    file:copy(filename:join(DataDir, "ssh_host_dsa_key.pub"), filename:join(System, "ssh_host_dsa_key.pub")),
-ct:log("DataDir ~p:~n ~p~n~nSystDir ~p:~n ~p~n~nUserDir ~p:~n ~p",[DataDir, file:list_dir(DataDir), System, file:list_dir(System), UserDir, file:list_dir(UserDir)]),
-    setup_dsa_known_host(DataDir, UserDir),
-    setup_dsa_auth_keys(DataDir, UserDir).
-    
-setup_rsa(DataDir, UserDir) ->
-    file:copy(filename:join(DataDir, "id_rsa"), filename:join(UserDir, "id_rsa")),
-    System = filename:join(UserDir, "system"),
-    file:make_dir(System),
-    file:copy(filename:join(DataDir, "ssh_host_rsa_key"), filename:join(System, "ssh_host_rsa_key")),
-    file:copy(filename:join(DataDir, "ssh_host_rsa_key.pub"), filename:join(System, "ssh_host_rsa_key.pub")),
-ct:log("DataDir ~p:~n ~p~n~nSystDir ~p:~n ~p~n~nUserDir ~p:~n ~p",[DataDir, file:list_dir(DataDir), System, file:list_dir(System), UserDir, file:list_dir(UserDir)]),
-    setup_rsa_known_host(DataDir, UserDir),
-    setup_rsa_auth_keys(DataDir, UserDir).
-
-setup_ecdsa(Size, DataDir, UserDir) ->
-    file:copy(filename:join(DataDir, "id_ecdsa"++Size), filename:join(UserDir, "id_ecdsa")),
-    System = filename:join(UserDir, "system"),
-    file:make_dir(System),
-    file:copy(filename:join(DataDir, "ssh_host_ecdsa_key"++Size), filename:join(System, "ssh_host_ecdsa_key")),
-    file:copy(filename:join(DataDir, "ssh_host_ecdsa_key"++Size++".pub"), filename:join(System, "ssh_host_ecdsa_key.pub")),
-ct:log("DataDir ~p:~n ~p~n~nSystDir ~p:~n ~p~n~nUserDir ~p:~n ~p",[DataDir, file:list_dir(DataDir), System, file:list_dir(System), UserDir, file:list_dir(UserDir)]),
-    setup_ecdsa_known_host(Size, System, UserDir),
-    setup_ecdsa_auth_keys(Size, DataDir, UserDir).
-
-clean_dsa(UserDir) ->
-    del_dirs(filename:join(UserDir, "system")),
-    file:delete(filename:join(UserDir,"id_dsa")),
-    file:delete(filename:join(UserDir,"known_hosts")),
-    file:delete(filename:join(UserDir,"authorized_keys")).
-
-clean_rsa(UserDir) ->
-    del_dirs(filename:join(UserDir, "system")),
-    file:delete(filename:join(UserDir,"id_rsa")),
-    file:delete(filename:join(UserDir,"known_hosts")),
-    file:delete(filename:join(UserDir,"authorized_keys")).
-
-setup_dsa_pass_pharse(DataDir, UserDir, Phrase) ->
-    {ok, KeyBin} = file:read_file(filename:join(DataDir, "id_dsa")),
-    setup_pass_pharse(KeyBin, filename:join(UserDir, "id_dsa"), Phrase),
-    System = filename:join(UserDir, "system"),
-    file:make_dir(System),
-    file:copy(filename:join(DataDir, "ssh_host_dsa_key"), filename:join(System, "ssh_host_dsa_key")),
-    file:copy(filename:join(DataDir, "ssh_host_dsa_key.pub"), filename:join(System, "ssh_host_dsa_key.pub")),
-    setup_dsa_known_host(DataDir, UserDir),
-    setup_dsa_auth_keys(DataDir, UserDir).
-
-setup_rsa_pass_pharse(DataDir, UserDir, Phrase) ->
-    {ok, KeyBin} = file:read_file(filename:join(DataDir, "id_rsa")),
-    setup_pass_pharse(KeyBin, filename:join(UserDir, "id_rsa"), Phrase),
-    System = filename:join(UserDir, "system"),
-    file:make_dir(System),
-    file:copy(filename:join(DataDir, "ssh_host_rsa_key"), filename:join(System, "ssh_host_rsa_key")),
-    file:copy(filename:join(DataDir, "ssh_host_rsa_key.pub"), filename:join(System, "ssh_host_rsa_key.pub")),
-    setup_rsa_known_host(DataDir, UserDir),
-    setup_rsa_auth_keys(DataDir, UserDir).
-
-setup_ecdsa_pass_phrase(Size, DataDir, UserDir, Phrase) ->
-    try
-        {ok, KeyBin} = 
-            case file:read_file(F=filename:join(DataDir, "id_ecdsa"++Size)) of
-                {error,E} ->
-                    ct:log("Failed (~p) to read ~p~nFiles: ~p", [E,F,file:list_dir(DataDir)]),
-                    file:read_file(filename:join(DataDir, "id_ecdsa"));
-                Other ->
-                    Other
-            end,
-        setup_pass_pharse(KeyBin, filename:join(UserDir, "id_ecdsa"), Phrase),
-        System = filename:join(UserDir, "system"),
-        file:make_dir(System),
-        file:copy(filename:join(DataDir, "ssh_host_ecdsa_key"++Size), filename:join(System, "ssh_host_ecdsa_key")),
-        file:copy(filename:join(DataDir, "ssh_host_ecdsa_key"++Size++".pub"), filename:join(System, "ssh_host_ecdsa_key.pub")),
-        setup_ecdsa_known_host(Size, System, UserDir),
-        setup_ecdsa_auth_keys(Size, DataDir, UserDir)
-    of 
-        _ -> true
-    catch
-        _:_ -> false
-    end.
-
-setup_pass_pharse(KeyBin, OutFile, Phrase) ->
-    [{KeyType, _,_} = Entry0] = public_key:pem_decode(KeyBin),
-    Key =  public_key:pem_entry_decode(Entry0),
-    Salt = crypto:strong_rand_bytes(8),
-    Entry = public_key:pem_entry_encode(KeyType, Key,
-					{{"DES-CBC", Salt}, Phrase}),
-    Pem = public_key:pem_encode([Entry]),
-    file:write_file(OutFile, Pem).
-
-setup_dsa_known_host(SystemDir, UserDir) ->
-    {ok, SshBin} = file:read_file(filename:join(SystemDir, "ssh_host_dsa_key.pub")),
-    [{Key, _}] = public_key:ssh_decode(SshBin, public_key),
-    setup_known_hosts(Key, UserDir).
-
-setup_rsa_known_host(SystemDir, UserDir) ->
-    {ok, SshBin} = file:read_file(filename:join(SystemDir, "ssh_host_rsa_key.pub")),
-    [{Key, _}] = public_key:ssh_decode(SshBin, public_key),
-    setup_known_hosts(Key, UserDir).
-
-setup_ecdsa_known_host(_Size, SystemDir, UserDir) ->
-    {ok, SshBin} = file:read_file(filename:join(SystemDir, "ssh_host_ecdsa_key.pub")),
-    [{Key, _}] = public_key:ssh_decode(SshBin, public_key),
-    setup_known_hosts(Key, UserDir).
-
-setup_known_hosts(Key, UserDir) ->
-    {ok, Hostname} = inet:gethostname(),
-    {ok, {A, B, C, D}} = inet:getaddr(Hostname, inet),
-    IP = lists:concat([A, ".", B, ".", C, ".", D]),
-    HostNames = [{hostnames,[Hostname, IP]}],
-    KnownHosts = [{Key, HostNames}],
-    KnownHostsEnc = public_key:ssh_encode(KnownHosts, known_hosts),
-    KHFile = filename:join(UserDir, "known_hosts"),
-    file:write_file(KHFile, KnownHostsEnc).
-
-setup_dsa_auth_keys(Dir, UserDir) ->
-    {ok, Pem} = file:read_file(filename:join(Dir, "id_dsa")),
-    DSA = public_key:pem_entry_decode(hd(public_key:pem_decode(Pem))),
-    PKey = DSA#'DSAPrivateKey'.y,
-    P = DSA#'DSAPrivateKey'.p,
-    Q = DSA#'DSAPrivateKey'.q,
-    G = DSA#'DSAPrivateKey'.g,
-    Dss = #'Dss-Parms'{p=P, q=Q, g=G},
-    setup_auth_keys([{{PKey, Dss}, [{comment, "Test"}]}], UserDir).
-
-setup_rsa_auth_keys(Dir, UserDir) ->
-    {ok, Pem} = file:read_file(filename:join(Dir, "id_rsa")),
-    RSA = public_key:pem_entry_decode(hd(public_key:pem_decode(Pem))),
-    #'RSAPrivateKey'{publicExponent = E, modulus = N} = RSA,
-    PKey = #'RSAPublicKey'{publicExponent = E, modulus = N},
-    setup_auth_keys([{ PKey, [{comment, "Test"}]}], UserDir).
-
-setup_ecdsa_auth_keys(Size, Dir, UserDir) ->
-    {ok, Pem} =
-        case file:read_file(F=filename:join(Dir, "id_ecdsa"++Size)) of
-            {error,E} ->
-                ct:log("Failed (~p) to read ~p~nFiles: ~p", [E,F,file:list_dir(Dir)]),
-                file:read_file(filename:join(Dir, "id_ecdsa"));
-            Other ->
-                Other
-        end,
-    ECDSA = public_key:pem_entry_decode(hd(public_key:pem_decode(Pem))),
-    #'ECPrivateKey'{publicKey = Q,
-		    parameters = Param = {namedCurve,_Id0}} = ECDSA,
-    PKey = #'ECPoint'{point = Q},
-    setup_auth_keys([{ {PKey,Param}, [{comment, "Test"}]}], UserDir).
-
-setup_auth_keys(Keys, Dir) ->
-    AuthKeys = public_key:ssh_encode(Keys, auth_keys),
-    AuthKeysFile = filename:join(Dir, "authorized_keys"),
-    ok = file:write_file(AuthKeysFile, AuthKeys),
-    AuthKeys.
-
-write_auth_keys(Keys, Dir) ->
-    AuthKeysFile = filename:join(Dir, "authorized_keys"),
-    file:write_file(AuthKeysFile, Keys).
-
 del_dirs(Dir) ->
+    del_dir_contents(Dir),
+    file:del_dir(Dir),
+    ok.
+
+
+del_dir_contents(Dir) ->
     case file:list_dir(Dir) of
-	{ok, []} ->
-	    file:del_dir(Dir);
-	{ok, Files} ->
-	    lists:foreach(fun(File) ->
-				  FullPath = filename:join(Dir,File),
-				  case filelib:is_dir(FullPath) of
-				      true ->
-					  del_dirs(FullPath),
-					  file:del_dir(FullPath);
-				      false ->
-					  file:delete(FullPath)
-				  end
-			  end, Files);
-	_ ->
-	    ok
+        {ok, Files} ->
+            do_del_files(Dir, Files);
+        _ ->
+            ok
     end.
 
-inet_port(Node) ->
-    {Port, Socket} = do_inet_port(Node),
-     rpc:call(Node, gen_tcp, close, [Socket]),
-     Port.
+do_del_files(Dir, Files) ->
+    lists:foreach(fun(File) ->
+                          FullPath = filename:join(Dir,File),
+                          case filelib:is_dir(FullPath) of
+                              true ->
+                                  del_dirs(FullPath);
+                              false ->
+                                  file:delete(FullPath)
+                          end
+                  end, Files).
 
-do_inet_port(Node) ->
-    {ok, Socket} = rpc:call(Node, gen_tcp, listen, [0, [{reuseaddr, true}]]),
-    {ok, Port} = rpc:call(Node, inet, port, [Socket]),
-    {Port, Socket}.
 
 openssh_sanity_check(Config) ->
     ssh:start(),
-    case ssh:connect("localhost", 22, [{password,""}]) of
+    case ssh:connect("localhost", 22, [{password,""},
+                                       {silently_accept_hosts, true},
+                                       {user_interaction, false}
+                                      ]) of
 	{ok, Pid} ->
 	    ssh:close(Pid),
 	    ssh:stop(),
@@ -581,34 +402,6 @@ openssh_sanity_check(Config) ->
 	    {skip, Str}
     end.
 
-openssh_supports(ClientOrServer, Tag, Alg) when ClientOrServer == sshc ;
-						ClientOrServer == sshd ->
-    SSH_algos = ssh_test_lib:default_algorithms(ClientOrServer),
-    L = proplists:get_value(Tag, SSH_algos, []),
-    lists:member(Alg, L) orelse 
-	lists:member(Alg, proplists:get_value(client2server, L, [])) orelse
-	lists:member(Alg, proplists:get_value(server2client, L, [])).
-
-%%--------------------------------------------------------------------
-%% Check if we have a "newer" ssh client that supports these test cases
-
-ssh_client_supports_Q() ->
-    0 == check_ssh_client_support2(
-	   ?MODULE:open_port({spawn, "ssh -Q cipher"})
-	  ).
-
-check_ssh_client_support2(P) ->
-    receive
-	{P, {data, _A}} ->
-	    check_ssh_client_support2(P);
-	{P, {exit_status, E}} ->
-            ct:log("~p:~p exit_status:~n~p",[?MODULE,?LINE,E]),
-	    E
-    after 5000 ->
-	    ct:log("Openssh command timed out ~n"),
-	    -1
-    end.
-
 %%%--------------------------------------------------------------------
 %%% Probe a server or a client about algorithm support
 
@@ -735,6 +528,15 @@ intersect_bi_dir([H={_,[A|_]}|T]) when is_atom(A) ->
 intersect_bi_dir([]) ->
     [].
     
+some_empty([]) ->
+    false;
+some_empty([{_,[]}|_]) ->
+    true;
+some_empty([{_,L}|T]) when is_atom(hd(L)) ->
+    some_empty(T);
+some_empty([{_,L}|T]) when is_tuple(hd(L)) ->
+    some_empty(L) orelse some_empty(T).
+
 
 sort_spec(L = [{_,_}|_] ) ->  [{Tag,sort_spec(Es)} || {Tag,Es} <- L];
 sort_spec(L) -> lists:usort(L).
@@ -945,7 +747,7 @@ expected_state(_) -> false.
 %%%----------------------------------------------------------------
 %%% Return a string with N random characters
 %%%
-random_chars(N) -> [crypto:rand_uniform($a,$z) || _<-lists:duplicate(N,x)].
+random_chars(N) -> [($a-1)+rand:uniform($z-$a) || _<-lists:duplicate(N,x)].
 
 
 create_random_dir(Config) ->
@@ -1029,3 +831,239 @@ ntoa(A) ->
         _:_ when is_list(A) -> A
     end.
     
+%%%----------------------------------------------------------------
+try_enable_fips_mode() ->
+    case crypto:info_fips() of
+        enabled ->
+            report("FIPS mode already enabled", ?LINE),
+            ok;
+        not_enabled ->
+            %% Erlang/crypto configured with --enable-fips
+            case crypto:enable_fips_mode(true) of
+		true ->
+                    %% and also the cryptolib is fips enabled
+                    report("FIPS mode enabled", ?LINE),
+		    enabled = crypto:info_fips(),
+		    ok;
+		false ->
+                    case is_cryptolib_fips_capable() of
+                        false ->
+                            report("No FIPS mode in cryptolib", ?LINE),
+                            {skip, "FIPS mode not supported in cryptolib"};
+                        true ->
+                            ct:fail("Failed to enable FIPS mode", [])
+                    end
+	    end;
+        not_supported ->
+            report("FIPS mode not supported by Erlang/OTP", ?LINE),
+            {skip, "FIPS mode not supported"}
+    end.
+
+is_cryptolib_fips_capable() ->
+    [{_,_,Inf}] = crypto:info_lib(),
+    nomatch =/= re:run(Inf, "(F|f)(I|i)(P|p)(S|s)").
+
+report(Comment, Line) ->
+    ct:comment(Comment),
+    ct:log("~p:~p  try_enable_fips_mode~n"
+           "crypto:info_lib() = ~p~n"
+           "crypto:info_fips() = ~p~n"
+           "crypto:supports() =~n~p~n", 
+           [?MODULE, Line,
+            crypto:info_lib(),
+            crypto:info_fips(),
+            crypto:supports()]).
+
+%%%----------------------------------------------------------------
+lc_name_in(Names) ->
+    case inet:gethostname() of
+        {ok,Name} ->
+            lists:member(string:to_lower(Name), Names);
+        Other ->
+            ct:log("~p:~p  inet:gethostname() returned ~p", [?MODULE,?LINE,Other]),
+            false
+    end.
+
+ptty_supported() -> not lc_name_in([]). %%["fobi"]).
+
+%%%----------------------------------------------------------------
+has_WSL() ->
+    os:getenv("WSLENV") =/= false. % " =/= false" =/= "== true" :)
+
+winpath_to_linuxpath(Path) ->
+    case {has_WSL(), Path} of
+        {true, [_,$:|WithoutWinInit]} ->
+            "/mnt/c" ++ WithoutWinInit;
+        _ ->
+            Path
+    end.
+    
+%%%----------------------------------------------------------------
+copy_recursive(Src, Dst) ->
+    {ok,S} = file:read_file_info(Src),
+    case S#file_info.type of
+        directory ->
+            %%ct:log("~p:~p copy dir  ~ts -> ~ts", [?MODULE,?LINE,Src,Dst]),
+            {ok,Names} = file:list_dir(Src),
+            mk_dir_path(Dst),
+            %%ct:log("~p:~p Names = ~p", [?MODULE,?LINE,Names]),
+            lists:foreach(fun(Name) ->
+                                  copy_recursive(filename:join(Src, Name),
+                                                 filename:join(Dst, Name))
+                          end, Names);
+        _ ->
+            %%ct:log("~p:~p copy file ~ts -> ~ts", [?MODULE,?LINE,Src,Dst]),
+            {ok,_NumBytesCopied} = file:copy(Src, Dst)
+    end.
+
+%%%----------------------------------------------------------------
+%% Make a directory even if parts of the path does not exist
+
+mk_dir_path(DirPath) ->
+    case file:make_dir(DirPath) of
+        {error,eexist} ->
+            %%ct:log("~p:~p dir exists ~ts", [?MODULE,?LINE,DirPath]),
+            ok;
+        {error,enoent} ->
+            %%ct:log("~p:~p try make dirname of ~ts", [?MODULE,?LINE,DirPath]),
+            case mk_dir_path( filename:dirname(DirPath) ) of
+                ok ->
+                    %%ct:log("~p:~p redo ~ts", [?MODULE,?LINE,DirPath]),
+                    file:make_dir(DirPath);
+                Error ->
+                    %%ct:log("~p:~p return Error ~p ~ts", [?MODULE,?LINE,Error,DirPath]),
+                    Error
+            end;
+        Other ->
+            %%ct:log("~p:~p return Other ~p ~ts", [?MODULE,?LINE,Other,DirPath]),
+            Other
+    end.
+
+%%%----------------------------------------------------------------
+%%% New
+
+setup_all_user_host_keys(Config) ->
+    DataDir = proplists:get_value(data_dir, Config),
+    PrivDir = proplists:get_value(priv_dir, Config),
+    setup_all_user_host_keys(DataDir, PrivDir).
+
+setup_all_user_host_keys(DataDir, PrivDir) ->
+    setup_all_user_host_keys(DataDir, PrivDir, filename:join(PrivDir,"system")).
+
+setup_all_user_host_keys(DataDir, UserDir, SysDir) ->
+    lists:foldl(fun(Alg, OkAlgs) ->
+                        try
+                            ok = ssh_test_lib:setup_user_key(Alg, DataDir, UserDir),
+                            ok = ssh_test_lib:setup_host_key(Alg, DataDir, SysDir)
+                        of
+                            ok -> [Alg|OkAlgs]
+                        catch
+                            error:{badmatch, {error,enoent}} ->
+                                OkAlgs;
+                            C:E:S ->
+                                ct:log("Exception in ~p:~p for alg ~p:  ~p:~p~n~p",
+                                       [?MODULE,?FUNCTION_NAME,Alg,C,E,S]),
+                                OkAlgs
+                        end
+                end, [], ssh_transport:supported_algorithms(public_key)).
+
+
+setup_all_host_keys(Config) ->
+    DataDir = proplists:get_value(data_dir, Config),
+    PrivDir = proplists:get_value(priv_dir, Config),
+    setup_all_host_keys(DataDir, filename:join(PrivDir,"system")).
+
+setup_all_host_keys(DataDir, SysDir) ->
+    lists:foldl(fun(Alg, OkAlgs) ->
+                        try
+                            ok = ssh_test_lib:setup_host_key(Alg, DataDir, SysDir)
+                        of
+                            ok -> [Alg|OkAlgs]
+                        catch
+                            error:{badmatch, {error,enoent}} ->
+                                OkAlgs;
+                            C:E:S ->
+                                ct:log("Exception in ~p:~p for alg ~p:  ~p:~p~n~p",
+                                       [?MODULE,?FUNCTION_NAME,Alg,C,E,S]),
+                                OkAlgs
+                        end
+                end, [], ssh_transport:supported_algorithms(public_key)).
+
+setup_user_key(SshAlg, DataDir, UserDir) ->
+    file:make_dir(UserDir),
+    %% Copy private user key to user's dir
+    {ok,_} = file:copy(filename:join(DataDir, file_base_name(user_src,SshAlg)),
+                       filename:join(UserDir, file_base_name(user,SshAlg))),
+    %% Setup authorized_keys in user's dir
+    {ok,Pub} = file:read_file(filename:join(DataDir, file_base_name(user_src,SshAlg)++".pub")),
+    ok = file:write_file(filename:join(UserDir, "authorized_keys"),
+                         io_lib:format("~n~s~n",[Pub]),
+                         [append]),
+    ?ct_log_show_file( filename:join(DataDir, file_base_name(user_src,SshAlg)++".pub") ),
+    ?ct_log_show_file( filename:join(UserDir, "authorized_keys") ),
+    ok.
+
+setup_host_key_create_dir(SshAlg, DataDir, BaseDir) ->
+    SysDir = filename:join(BaseDir,"system"),
+    ct:log("~p:~p  SshAlg=~p~nDataDir = ~p~nBaseDir = ~p~nSysDir = ~p",[?MODULE,?LINE,SshAlg, DataDir, BaseDir,SysDir]),
+    file:make_dir(SysDir),
+    setup_host_key(SshAlg, DataDir, SysDir),
+    SysDir.
+
+setup_host_key(SshAlg, DataDir, SysDir) ->
+    mk_dir_path(SysDir),
+    %% Copy private host key to system's dir
+    {ok,_} = file:copy(filename:join(DataDir, file_base_name(system_src,SshAlg)),
+                       filename:join(SysDir,  file_base_name(system,SshAlg))),
+    ?ct_log_show_file( filename:join(SysDir,  file_base_name(system,SshAlg)) ),
+    ok.
+
+setup_known_host(SshAlg, DataDir, UserDir) ->
+    {ok,Pub} = file:read_file(filename:join(DataDir, file_base_name(system_src,SshAlg)++".pub")),
+    S = lists:join(" ", lists:reverse(tl(lists:reverse(string:tokens(binary_to_list(Pub), " "))))),
+    ok = file:write_file(filename:join(UserDir, "known_hosts"),
+                         io_lib:format("~p~n",[S])),
+    ?ct_log_show_file( filename:join(UserDir, "known_hosts") ),
+    ok.
+
+
+get_addr_str() ->
+    {ok, Hostname} = inet:gethostname(),
+    {ok, {A, B, C, D}} = inet:getaddr(Hostname, inet),
+    IP = lists:concat([A, ".", B, ".", C, ".", D]),
+    lists:concat([Hostname,",",IP]).
+
+
+file_base_name(user,   'ecdsa-sha2-nistp256') -> "id_ecdsa";
+file_base_name(user,   'ecdsa-sha2-nistp384') -> "id_ecdsa";
+file_base_name(user,   'ecdsa-sha2-nistp521') -> "id_ecdsa";
+file_base_name(user,   'rsa-sha2-256'       ) -> "id_rsa";
+file_base_name(user,   'rsa-sha2-384'       ) -> "id_rsa";
+file_base_name(user,   'rsa-sha2-512'       ) -> "id_rsa";
+file_base_name(user,   'ssh-dss'            ) -> "id_dsa";
+file_base_name(user,   'ssh-ed25519'        ) -> "id_ed25519";
+file_base_name(user,   'ssh-ed448'          ) -> "id_ed448";
+file_base_name(user,   'ssh-rsa'            ) -> "id_rsa";
+
+file_base_name(user_src, 'ecdsa-sha2-nistp256') -> "id_ecdsa256";
+file_base_name(user_src, 'ecdsa-sha2-nistp384') -> "id_ecdsa384";
+file_base_name(user_src, 'ecdsa-sha2-nistp521') -> "id_ecdsa521";
+file_base_name(user_src, Alg) -> file_base_name(user, Alg);
+
+file_base_name(system, 'ecdsa-sha2-nistp256') -> "ssh_host_ecdsa_key";
+file_base_name(system, 'ecdsa-sha2-nistp384') -> "ssh_host_ecdsa_key";
+file_base_name(system, 'ecdsa-sha2-nistp521') -> "ssh_host_ecdsa_key";
+file_base_name(system, 'rsa-sha2-256'       ) -> "ssh_host_rsa_key";
+file_base_name(system, 'rsa-sha2-384'       ) -> "ssh_host_rsa_key";
+file_base_name(system, 'rsa-sha2-512'       ) -> "ssh_host_rsa_key";
+file_base_name(system, 'ssh-dss'            ) -> "ssh_host_dsa_key";
+file_base_name(system, 'ssh-ed25519'        ) -> "ssh_host_ed25519_key";
+file_base_name(system, 'ssh-ed448'          ) -> "ssh_host_ed448_key";
+file_base_name(system, 'ssh-rsa'            ) -> "ssh_host_rsa_key";
+
+file_base_name(system_src, 'ecdsa-sha2-nistp256') -> "ssh_host_ecdsa_key256";
+file_base_name(system_src, 'ecdsa-sha2-nistp384') -> "ssh_host_ecdsa_key384";
+file_base_name(system_src, 'ecdsa-sha2-nistp521') -> "ssh_host_ecdsa_key521";
+file_base_name(system_src, Alg) -> file_base_name(system, Alg).
+
+%%%----------------------------------------------------------------
