@@ -37,24 +37,32 @@ all() ->
      {group, 'tlsv1.2'},
      {group, 'tlsv1.1'},
      {group, 'tlsv1'},
-     {group, 'sslv3'},
      {group, 'dtlsv1.2'},
      {group, 'dtlsv1'}
     ].
 
 groups() ->
     [
-     {'tlsv1.3', [], ((gen_api_tests() ++ tls13_group() ++ handshake_paus_tests()) -- [dh_params, honor_server_cipher_order, honor_client_cipher_order,
-                                        new_options_in_handshake])
+     {'tlsv1.3', [], ((gen_api_tests() ++ tls13_group() ++
+                           handshake_paus_tests()) --
+                          [dh_params,
+                           honor_server_cipher_order,
+                           honor_client_cipher_order,
+                           new_options_in_handshake,
+                           handshake_continue_tls13_client,
+                           invalid_options])
       ++ (since_1_2() -- [conf_signature_algs])},
      {'tlsv1.2', [],  gen_api_tests() ++ since_1_2() ++ handshake_paus_tests() ++ pre_1_3()},
      {'tlsv1.1', [],  gen_api_tests() ++ handshake_paus_tests() ++ pre_1_3()},
      {'tlsv1', [],  gen_api_tests() ++ handshake_paus_tests() ++ pre_1_3() ++ beast_mitigation_test()},
-     {'sslv3', [],  (gen_api_tests() -- [new_options_in_handshake]) ++ beast_mitigation_test() ++ pre_1_3()},
-     {'dtlsv1.2', [], (gen_api_tests() -- [invalid_keyfile, invalid_certfile, invalid_cacertfile,
-                                         invalid_options, new_options_in_handshake])  ++ handshake_paus_tests() ++ pre_1_3()},
-     {'dtlsv1', [],  (gen_api_tests() -- [invalid_keyfile, invalid_certfile, invalid_cacertfile,
-                                         invalid_options, new_options_in_handshake]) ++ handshake_paus_tests() ++ pre_1_3()}
+     {'dtlsv1.2', [], (gen_api_tests() --
+                           [invalid_keyfile, invalid_certfile, invalid_cacertfile,
+                            invalid_options, new_options_in_handshake])  ++
+          handshake_paus_tests() -- [handshake_continue_tls13_client] ++ pre_1_3()},
+     {'dtlsv1', [],  (gen_api_tests() --
+                          [invalid_keyfile, invalid_certfile, invalid_cacertfile,
+                           invalid_options, new_options_in_handshake]) ++
+          handshake_paus_tests() -- [handshake_continue_tls13_client] ++ pre_1_3()}
     ].
 
 since_1_2() ->
@@ -65,7 +73,8 @@ since_1_2() ->
 
 pre_1_3() ->
     [
-     default_reject_anonymous
+     default_reject_anonymous,
+     connection_information_with_srp
     ].
 gen_api_tests() ->
     [
@@ -102,7 +111,10 @@ gen_api_tests() ->
      invalid_cacertfile,
      invalid_keyfile,
      options_not_proplist,
-     invalid_options
+     invalid_options,
+     cb_info,
+     log_alert,
+     getstat
     ].
 
 handshake_paus_tests() ->
@@ -110,7 +122,8 @@ handshake_paus_tests() ->
      handshake_continue, 
      handshake_continue_timeout, 
      hello_client_cancel,
-     hello_server_cancel
+     hello_server_cancel,
+     handshake_continue_tls13_client
     ].
 
 %% Only relevant for SSL 3.0 and TLS 1.1
@@ -127,10 +140,16 @@ tls13_group() ->
     [
      supported_groups,
      honor_server_cipher_order_tls13,
-     honor_client_cipher_order_tls13
+     honor_client_cipher_order_tls13,
+     client_options_negative_version_gap,
+     client_options_negative_dependency_version,
+     client_options_negative_dependency_stateless,
+     client_options_negative_dependency_role,
+     server_options_negative_version_gap,
+     server_options_negative_dependency_role,
+     invalid_options_tls13
     ].
 
-
 init_per_suite(Config0) ->
     catch crypto:stop(),
     try crypto:start() of
@@ -146,29 +165,17 @@ end_per_suite(_Config) ->
     application:unload(ssl),
     application:stop(crypto).
 
-
-init_per_group(GroupName, Config) ->
-    case ssl_test_lib:is_tls_version(GroupName) of
-	true ->
-	    case ssl_test_lib:sufficient_crypto_support(GroupName) of
-		true ->
-		    [{client_type, erlang},
-                     {server_type, erlang} | ssl_test_lib:init_tls_version(GroupName, Config)];
-		false ->
-		    {skip, "Missing crypto support"}
-	    end;
-	_ ->
-	    ssl:start(),
-	    Config
+init_per_group(GroupName, Config0) ->
+    case ssl_test_lib:init_per_group(GroupName, Config0) of
+        {skip, _} = Skip ->
+            Skip;
+        Config ->
+            [{client_type, erlang},
+             {server_type, erlang}|Config]
     end.
 
 end_per_group(GroupName, Config) ->
-    case ssl_test_lib:is_tls_version(GroupName) of
-        true ->
-            ssl_test_lib:clean_tls_version(Config);
-        false ->
-            Config
-    end.
+    ssl_test_lib:end_per_group(GroupName, Config).
 
 init_per_testcase(prf, Config) ->
     ssl_test_lib:ct_log_supported_protocol_versions(Config),
@@ -187,6 +194,23 @@ init_per_testcase(prf, Config) ->
          {md5sha, <<63,136,3,217,205,123,200,177,251,211,17,229,132,4,173,80>>}],
     TestPlan = prf_create_plan([Version], PRFS, ExpectedPrfResults),
     [{prf_test_plan, TestPlan} | Config];
+init_per_testcase(handshake_continue_tls13_client, Config) ->
+    case ssl_test_lib:sufficient_crypto_support('tlsv1.3') of
+        true ->
+            ssl_test_lib:ct_log_supported_protocol_versions(Config),
+            ct:timetrap({seconds, 10}),
+            Config;
+        false ->
+            {skip, "Missing crypto support: TLS 1.3 not supported"}
+    end;
+init_per_testcase(connection_information_with_srp, Config) ->
+    PKAlg = proplists:get_value(public_keys, crypto:supports()),
+    case lists:member(srp, PKAlg) of
+        true ->
+            Config;
+        false ->
+            {skip, "Missing SRP crypto support"}
+    end;
 init_per_testcase(_TestCase, Config) ->
     ssl_test_lib:ct_log_supported_protocol_versions(Config),
     ct:timetrap({seconds, 10}),
@@ -294,6 +318,56 @@ connection_information(Config) when is_list(Config) ->
     
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
+
+%%--------------------------------------------------------------------
+connection_information_with_srp() ->
+    [{doc,"Test the result of API function ssl:connection_information/1"
+          "includes srp_username."}].
+connection_information_with_srp(Config) when is_list(Config) ->
+    run_conn_info_srp_test(srp_anon, 'aes_128_cbc', Config).
+
+run_conn_info_srp_test(Kex, Cipher, Config) ->
+    Version = ssl_test_lib:protocol_version(Config),
+    TestCiphers = ssl_test_lib:test_ciphers(Kex, Cipher, Version),
+
+    case TestCiphers of
+        [] ->
+            {skip, {not_sup, Kex, Cipher, Version}};
+        [TestCipher | _T] ->
+            do_run_conn_info_srp_test(TestCipher, Version, Config)
+    end.
+
+do_run_conn_info_srp_test(ErlangCipherSuite, Version, Config) ->
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+
+    SOpts = [{user_lookup_fun, {fun ssl_test_lib:user_lookup/3, undefined}}],
+    COpts = [{srp_identity, {"Test-User", "secret"}}],
+
+    ServerOpts = ssl_test_lib:ssl_options(SOpts, Config),
+    ClientOpts = ssl_test_lib:ssl_options(COpts, Config),
+
+    ct:log("Erlang Cipher Suite is: ~p~n", [ErlangCipherSuite]),
+
+    Server = ssl_test_lib:start_server([
+        {node, ServerNode}, {port, 0},
+		{from, self()},
+        {mfa, {?MODULE, check_srp_in_connection_information, [<<"Test-User">>, server]}},
+        {options, [{versions, [Version]}, {ciphers, [ErlangCipherSuite]} |
+                   ServerOpts]}]),
+    Port = ssl_test_lib:inet_port(Server),
+    Client = ssl_test_lib:start_client(
+        [{node, ClientNode}, {port, Port},
+		 {host, Hostname},
+		 {from, self()},
+		 {mfa, {?MODULE, check_srp_in_connection_information, [<<"Test-User">>, client]}},
+		 {options, [{versions, [Version]}, {ciphers, [ErlangCipherSuite]} |
+                    ClientOpts]}]),
+
+    ssl_test_lib:check_result(Server, ok, Client, ok),
+
+    ssl_test_lib:close(Server),
+    ssl_test_lib:close(Client).
+
 %%--------------------------------------------------------------------
 
 secret_connection_info() ->
@@ -435,7 +509,8 @@ handshake_continue(Config) when is_list(Config) ->
     Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0}, 
 					{from, self()}, 
 					{mfa, {ssl_test_lib, send_recv_result_active, []}},
-                                        {options, ssl_test_lib:ssl_options([{reuseaddr, true}, 
+                                        {options, ssl_test_lib:ssl_options([{reuseaddr, true},
+                                                                            {log_level, debug},
                                                                             {verify, verify_peer},
                                                                             {handshake, hello} | ServerOpts
                                                                            ], 
@@ -460,6 +535,70 @@ handshake_continue(Config) when is_list(Config) ->
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
 
+%%--------------------------------------------------------------------
+handshake_continue_tls13_client() ->
+    [{doc, "Test API function ssl:handshake_continue/3 with fixed TLS 1.3 client"}].
+handshake_continue_tls13_client(Config) when is_list(Config) ->
+    ClientOpts = ssl_test_lib:ssl_options(client_rsa_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_rsa_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+
+    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0},
+					{from, self()},
+					{mfa, {ssl_test_lib, send_recv_result_active, []}},
+                                        {options, ssl_test_lib:ssl_options([{reuseaddr, true},
+                                                                            {log_level, debug},
+                                                                            {verify, verify_peer},
+                                                                            {handshake, hello} | ServerOpts
+                                                                           ],
+                                                                           Config)},
+                                        {continue_options, proplists:delete(reuseaddr, ServerOpts)}
+                                       ]),
+
+    Port = ssl_test_lib:inet_port(Server),
+
+    DummyTicket =
+        <<131,116,0,0,0,5,100,0,4,104,107,100,102,100,0,6,115,104,97,51,
+          56,52,100,0,3,112,115,107,109,0,0,0,48,150,90,38,127,26,12,5,
+          228,180,235,229,214,215,27,236,149,182,82,14,140,50,81,0,150,
+          248,152,180,193,207,80,52,107,196,200,2,77,4,96,140,65,239,205,
+          224,125,129,179,147,103,100,0,3,115,110,105,107,0,25,112,114,
+          111,117,100,102,111,111,116,46,111,116,112,46,101,114,105,99,
+          115,115,111,110,46,115,101,100,0,6,116,105,99,107,101,116,104,6,
+          100,0,18,110,101,119,95,115,101,115,115,105,111,110,95,116,105,
+          99,107,101,116,98,0,0,28,32,98,127,110,83,249,109,0,0,0,8,0,0,0,
+          0,0,0,0,5,109,0,0,0,113,112,154,74,26,27,0,111,147,51,110,216,
+          43,45,4,100,215,152,195,118,96,22,34,1,184,170,42,166,238,109,
+          187,138,196,147,102,205,116,83,241,174,227,232,156,148,60,153,3,
+          175,128,115,192,36,103,191,239,58,222,192,172,190,239,92,8,131,
+          195,0,217,187,222,143,104,6,86,53,93,27,218,198,205,138,223,202,
+          11,55,168,104,6,219,228,217,157,37,52,205,252,165,135,167,116,
+          216,172,231,222,189,84,97,0,8,106,108,88,47,114,48,116,0,0,0,0,
+          100,0,9,116,105,109,101,115,116,97,109,112,98,93,205,0,44>>,
+
+    %% Send dummy session ticket to trigger sending of pre_shared_key and
+    %% psk_key_exchange_modes extensions.
+    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port},
+					{host, Hostname},
+                                        {from, self()},
+                                        {mfa, {ssl_test_lib, send_recv_result_active, []}},
+                                        {options, ssl_test_lib:ssl_options([{handshake, hello},
+                                                                            {session_tickets, manual},
+                                                                            {use_ticket, [DummyTicket]},
+                                                                            {versions, ['tlsv1.3',
+                                                                                        'tlsv1.2',
+                                                                                        'tlsv1.1',
+                                                                                        'tlsv1'
+                                                                                       ]},
+                                                                            {verify, verify_peer} | ClientOpts
+                                                                           ],
+                                                                           Config)},
+                                        {continue_options,  proplists:delete(reuseaddr, ClientOpts)}]),
+
+    ssl_test_lib:check_result(Server, ok, Client, ok),
+
+    ssl_test_lib:close(Server),
+    ssl_test_lib:close(Client).
 
 %%------------------------------------------------------------------
 handshake_continue_timeout() ->
@@ -1479,7 +1618,6 @@ invalid_options(Config) when is_list(Config) ->
           {verify, 4}, 
           {verify_fun, function},
           {fail_if_no_peer_cert, 0}, 
-          {verify_client_once, 1},
           {depth, four}, 
           {certfile, 'cert.pem'}, 
           {keyfile,'key.pem' }, 
@@ -1497,6 +1635,20 @@ invalid_options(Config) when is_list(Config) ->
           {active, trice},
           {key, 'key.pem' }],
 
+    TestOpts2 =
+        [{[{anti_replay, '10k'}],
+          %% anti_replay is a server only option but tested with client
+          %% for simplicity
+          {options,dependency,{anti_replay,{versions,['tlsv1.3']}}}},
+         {[{supported_groups, []}],
+          {options,dependency,{supported_groups,{versions,['tlsv1.3']}}}},
+         {[{use_ticket, [<<1,2,3,4>>]}],
+          {options,dependency,{use_ticket,{versions,['tlsv1.3']}}}},
+         {[{verify, verify_none}, {fail_if_no_peer_cert, true}],
+          {options, incompatible,
+                   {verify, verify_none},
+                   {fail_if_no_peer_cert, true}}}],
+
     [begin
 	 Server =
 	     ssl_test_lib:start_server_error([{node, ServerNode}, {port, 0},
@@ -1510,7 +1662,13 @@ invalid_options(Config) when is_list(Config) ->
 	 Check(Client, Server, TestOpt),
 	 ok
      end || TestOpt <- TestOpts],
+
+    [begin
+         start_client_negative(Config, TestOpt, ErrorMsg),
+         ok
+     end || {TestOpt, ErrorMsg} <- TestOpts2],
     ok.
+
 %%-------------------------------------------------------------------
 
 default_reject_anonymous()->
@@ -1537,6 +1695,60 @@ default_reject_anonymous(Config) when is_list(Config) ->
 
     ssl_test_lib:check_server_alert(Server, Client, insufficient_security).
 
+%%-------------------------------------------------------------------
+cb_info() ->
+    [{doc,"Test that we can set cb_info."}].
+
+cb_info(Config) when is_list(Config) ->
+    ClientOpts = ssl_test_lib:ssl_options(client_rsa_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_rsa_opts, Config),
+    Protocol = proplists:get_value(protocol, Config, tls),
+    CbInfo =
+        case Protocol of
+            tls ->
+                {cb_info, {gen_tcp, tcp, tcp_closed, tcp_error}};
+            dtls ->
+                {cb_info, {gen_udp, udp, udp_closed, udp_error}}
+        end,
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0},
+                                        {from, self()},
+                                        {mfa, {ssl_test_lib, send_recv_result_active, []}},
+                                        {options,  [CbInfo | ServerOpts]}]),
+    Port = ssl_test_lib:inet_port(Server),
+
+    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port},
+                                        {host, Hostname},
+                                        {from, self()},
+                                        {mfa, {ssl_test_lib, send_recv_result_active, []}},
+                                        {options, [CbInfo | ClientOpts]}]),
+
+    ssl_test_lib:check_result(Server, ok, Client, ok).
+
+%%-------------------------------------------------------------------
+log_alert() ->
+    [{doc,"Test that we can set log_alert."}].
+
+log_alert(Config) when is_list(Config) ->
+    ClientOpts = ssl_test_lib:ssl_options(client_rsa_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_rsa_opts, Config),
+
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0},
+                                        {from, self()},
+                                        {mfa, {ssl_test_lib, send_recv_result_active, []}},
+                                        {options,  [{log_alert, false} | ServerOpts]}]),
+    Port = ssl_test_lib:inet_port(Server),
+
+    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port},
+                                        {host, Hostname},
+                                        {from, self()},
+                                        {mfa, {ssl_test_lib, send_recv_result_active, []}},
+                                        {options, [{log_alert, false} | ClientOpts]}]),
+
+    ssl_test_lib:check_result(Server, ok, Client, ok).
+
+
 %%-------------------------------------------------------------------
 %% Note that these test only test that the options are valid to set. As application data
 %% is a stream you can not test that the send acctually splits it up as when it arrives
@@ -1620,6 +1832,60 @@ honor_client_cipher_order_tls13(Config) when is_list(Config) ->
                                                                       mac => aead,
                                                                       prf => sha384}).
 
+%%--------------------------------------------------------------------
+client_options_negative_version_gap() ->
+    [{doc,"Test client options with faulty version gap."}].
+client_options_negative_version_gap(Config) when is_list(Config) ->
+    start_client_negative(Config, [{versions, ['tlsv1', 'tlsv1.3']}],
+                          {options, missing_version, 
+                           {'tlsv1.2', {versions,[tlsv1, 'tlsv1.3']}}}).
+
+%%--------------------------------------------------------------------
+client_options_negative_dependency_version() ->
+    [{doc,"Test client options with faulty version dependency."}].
+client_options_negative_dependency_version(Config) when is_list(Config) ->
+    start_client_negative(Config, [{versions, ['tlsv1.1', 'tlsv1.2']},
+                                   {session_tickets, manual}],
+                          {options,dependency,
+                           {session_tickets,{versions,['tlsv1.3']}}}).
+
+%%--------------------------------------------------------------------
+client_options_negative_dependency_stateless() ->
+    [{doc,"Test client options with faulty 'session_tickets' option."}].
+client_options_negative_dependency_stateless(Config) when is_list(Config) ->
+    start_client_negative(Config, [{versions, ['tlsv1.2', 'tlsv1.3']},
+                                   {anti_replay, '10k'},
+                                   {session_tickets, manual}],
+                          {options,dependency,
+                           {anti_replay,{session_tickets,[stateless]}}}).
+
+
+%%--------------------------------------------------------------------
+client_options_negative_dependency_role() ->
+    [{doc,"Test client options with faulty role."}].
+client_options_negative_dependency_role(Config) when is_list(Config) ->
+    start_client_negative(Config, [{versions, ['tlsv1.2', 'tlsv1.3']},
+                                   {session_tickets, stateless}],
+                          {options,role,
+                           {session_tickets,{stateless,{client,[disabled,manual,auto]}}}}).
+
+%%--------------------------------------------------------------------
+server_options_negative_version_gap() ->
+    [{doc,"Test server options with faulty version gap."}].
+server_options_negative_version_gap(Config) when is_list(Config) ->
+    start_server_negative(Config, [{versions, ['tlsv1', 'tlsv1.3']}],
+                          {options, missing_version,
+                           {'tlsv1.2', {versions,[tlsv1, 'tlsv1.3']}}}).
+
+%%--------------------------------------------------------------------
+server_options_negative_dependency_role() ->
+    [{doc,"Test server options with faulty role."}].
+server_options_negative_dependency_role(Config) when is_list(Config) ->
+    start_server_negative(Config, [{versions, ['tlsv1.2', 'tlsv1.3']},
+                                   {session_tickets, manual}],
+                          {options,role,
+                           {session_tickets,{manual,{server,[disabled,stateful,stateless]}}}}).
+
 %%--------------------------------------------------------------------
 honor_server_cipher_order_tls13() ->
     [{doc,"Test API honor server cipher order in TLS 1.3."}].
@@ -1644,6 +1910,128 @@ honor_server_cipher_order_tls13(Config) when is_list(Config) ->
                                                                      cipher => aes_128_gcm,
                                                                      mac => aead,
                                                                      prf => sha256}).
+%%--------------------------------------------------------------------
+getstat() ->
+    [{doc, "Test that you use ssl:getstat on a TLS socket"}].
+
+getstat(Config) when is_list(Config) ->    
+    ServerOpts = ssl_test_lib:ssl_options(server_rsa_opts, Config),
+    ClientOpts = ssl_test_lib:ssl_options(client_rsa_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    
+    Port0 = ssl_test_lib:inet_port(ServerNode),
+    {ok, ListenSocket} = ssl:listen(Port0, [ServerOpts]),
+    {ok, _} = ssl:getstat(ListenSocket),
+    ssl:close(ListenSocket),
+
+    Server = ssl_test_lib:start_server([{node, ClientNode}, {port, 0},
+					{from, self()},
+                                        {mfa, {?MODULE, ssl_getstat, []}},
+                                        {options, ServerOpts}]),
+    Port = ssl_test_lib:inet_port(Server),
+    Client = ssl_test_lib:start_client([{node, ServerNode}, {port, Port},
+					{host, Hostname},
+                                        {from, self()},
+                                        {mfa, {?MODULE, ssl_getstat, []}},
+                                        {options, ClientOpts}]),
+    ssl_test_lib:check_result(Server, ok, Client, ok).
+
+invalid_options_tls13() ->
+    [{doc, "Test invalid options with TLS 1.3"}].
+invalid_options_tls13(Config) when is_list(Config) ->
+    TestOpts =
+        [{{beast_mitigation, one_n_minus_one},
+          {options, dependency,
+           {beast_mitigation,{versions,[tlsv1]}}},
+          common},
+
+         {{next_protocols_advertised, [<<"http/1.1">>]},
+          {options, dependency,
+           {next_protocols_advertised,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          server},
+
+         {{client_preferred_next_protocols,
+           {client, [<<"http/1.1">>]}},
+          {options, dependency,
+           {client_preferred_next_protocols,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          client},
+
+         {{client_renegotiation, false},
+          {options, dependency,
+           {client_renegotiation,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          server
+         },
+
+         {{padding_check, false},
+          {options, dependency,
+           {padding_check,{versions,[tlsv1]}}},
+          common},
+
+         {{psk_identity, "Test-User"},
+          {options, dependency,
+           {psk_identity,{versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          common},
+
+         {{user_lookup_fun,
+           {fun ssl_test_lib:user_lookup/3, <<1,2,3>>}},
+          {options, dependency,
+           {user_lookup_fun,{versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          common},
+
+         {{reuse_session, fun(_,_,_,_) -> false end},
+          {options, dependency,
+           {reuse_session,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          server},
+
+         {{reuse_session, <<1,2,3,4>>},
+          {options, dependency,
+           {reuse_session,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          client},
+
+         {{reuse_sessions, true},
+          {options, dependency,
+           {reuse_sessions,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          common},
+
+         {{secure_renegotiate, false},
+          {options, dependency,
+           {secure_renegotiate,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          common},
+
+         {{srp_identity, false},
+          {options, dependency,
+           {srp_identity,
+            {versions,[tlsv1,'tlsv1.1','tlsv1.2']}}},
+          client}
+        ],
+
+    Fun = fun(Option, ErrorMsg, Type) ->
+                  case Type of
+                      server ->
+                          start_server_negative(Config,
+                                                [Option,{versions, ['tlsv1.3']}],
+                                                ErrorMsg);
+                      client ->
+                          start_client_negative(Config,
+                                                [Option,{versions, ['tlsv1.3']}],
+                                                ErrorMsg);
+                      common ->
+                          start_server_negative(Config,
+                                                [Option,{versions, ['tlsv1.3']}],
+                                                ErrorMsg),
+                          start_client_negative(Config,
+                                                [Option,{versions, ['tlsv1.3']}],
+                                                ErrorMsg)
+                  end
+          end,
+    [Fun(Option, ErrorMsg, Type) || {Option, ErrorMsg, Type} <- TestOpts].
 
 
 %%--------------------------------------------------------------------
@@ -1667,7 +2055,18 @@ secret_connection_info_result(Socket) ->
     {ok, [{protocol, Protocol}]} = ssl:connection_information(Socket, [protocol]),
     {ok, ConnInfo} = ssl:connection_information(Socket, [client_random, server_random, master_secret]),
     check_connection_info(Protocol, ConnInfo).
-  
+
+check_srp_in_connection_information(_Socket, _Username, client) ->
+    ok;
+check_srp_in_connection_information(Socket, Username, server) ->
+    {ok, Info} = ssl:connection_information(Socket),
+    ct:log("Info ~p~n", [Info]),
+    case proplists:get_value(srp_username, Info, not_found) of
+        Username ->
+	        ok;
+        not_found ->
+            ct:fail(srp_username_not_found)
+    end.
 
 %% In TLS 1.3 the master_secret field is used to store multiple secrets from the key schedule and it is a tuple.
 %% client_random and server_random are not used in the TLS 1.3 key schedule.
@@ -1958,6 +2357,28 @@ honor_cipher_order(Config, Honor, ServerCiphers, ClientCiphers, Expected) ->
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
 
+start_client_negative(Config, Options, Error) ->
+    ClientOpts = ssl_test_lib:ssl_options(client_rsa_opts, Config),
+    {ClientNode, _, Hostname} = ssl_test_lib:run_where(Config),
+    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, 0},
+					{host, Hostname},
+					{from, self()},
+					{mfa, {?MODULE, connection_info_result, []}},
+					{options, Options ++ ClientOpts}]),
+    ct:pal("Actual: ~p~nExpected: ~p", [Client, {connect_failed, Error}]),
+    {connect_failed, Error} = Client.
+
+start_server_negative(Config, Options, Error) ->
+    ServerOpts = ssl_test_lib:ssl_options(server_rsa_opts, Config),
+    {_, ServerNode, _} = ssl_test_lib:run_where(Config),
+    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0},
+					{from, self()},
+					{mfa, {?MODULE, connection_info_result, []}},
+					{options,  Options ++ ServerOpts}]),
+
+    ct:pal("Actual: ~p~nExpected: ~p", [Server,Error]),
+    Error = Server.
+
 connection_info_result(Socket) ->
     {ok, Info} = ssl:connection_information(Socket, [protocol, selected_cipher_suite]),
     {ok, {proplists:get_value(protocol, Info), proplists:get_value(selected_cipher_suite, Info)}}.
@@ -1976,3 +2397,18 @@ der_input_opts(Opts) ->
 			  CaCert
 		  end, ssl_test_lib:pem_to_der(CaCertsfile)),
     {Cert, {Asn1Type, Key}, CaCerts, DHParams}.
+
+ssl_getstat(Socket) ->
+    ssl:send(Socket, "From Erlang to Erlang"),
+    {ok, Stats} = ssl:getstat(Socket),
+    List = lists:dropwhile(fun({_, 0}) ->
+                                   true;
+                              ({_, _}) ->
+                                   false
+                           end, Stats),
+    case List of
+        [] ->
+            nok;
+        _  ->
+            ok
+    end.
