@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2008-2020. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -24,54 +24,64 @@
 -compile(export_all).
 
 -include_lib("common_test/include/ct.hrl").
-
--define(TIMEOUT, 600000).
+-define(TIMEOUT, {seconds, 20}).
+-define(TIMEOUT_LONG, {seconds, 80}).
 
 %%--------------------------------------------------------------------
 %% Common Test interface functions -----------------------------------
 %%--------------------------------------------------------------------
 all() -> 
     [
+     {group, 'tlsv1.3'},
      {group, 'tlsv1.2'},
      {group, 'tlsv1.1'},
-     {group, 'tlsv1'},
-     {group, 'sslv3'}
+     {group, 'tlsv1'}    
     ].
 
 groups() ->
     [
+     {'tlsv1.3', [], payload_tests()},
      {'tlsv1.2', [], payload_tests()},
      {'tlsv1.1', [], payload_tests()},
-     {'tlsv1', [], payload_tests()},
-     {'sslv3', [], payload_tests()}
+     {'tlsv1', [], payload_tests()}
     ].
 
 payload_tests() ->
     [server_echos_passive_small,
+     server_echos_passive_chunk_small,
      server_echos_active_once_small,
      server_echos_active_small,
      client_echos_passive_small,
+     client_echos_passive_chunk_small,
      client_echos_active_once_small,
      client_echos_active_small,
      server_echos_passive_big,
+     server_echos_passive_chunk_big,
      server_echos_active_once_big,
      server_echos_active_big,
      client_echos_passive_big,
+     client_echos_passive_chunk_big,
      client_echos_active_once_big,
      client_echos_active_big,
      server_echos_passive_huge,
+     server_echos_passive_chunk_huge,
      server_echos_active_once_huge,
      server_echos_active_huge,
      client_echos_passive_huge,
+     client_echos_passive_chunk_huge,
      client_echos_active_once_huge,
-     client_echos_active_huge].
+     client_echos_active_huge,
+     client_active_once_server_close].
 
 init_per_suite(Config) ->
     catch crypto:stop(),
     try crypto:start() of
 	ok ->
 	    ssl_test_lib:clean_start(),
-	    {ok, _} = make_certs:all(proplists:get_value(data_dir, Config), proplists:get_value(priv_dir, Config)),
+	    {ok, _} =
+                make_certs:all(
+                  proplists:get_value(data_dir, Config),
+                  proplists:get_value(priv_dir, Config)),
 	    ssl_test_lib:cert_options(Config)
     catch _:_  ->
 	    {skip, "Crypto did not start"}
@@ -103,31 +113,37 @@ end_per_group(GroupName, Config) ->
             Config
     end.
 
-init_per_testcase(TestCase, Config) when TestCase == server_echos_passive_huge;
-					 TestCase == server_echos_active_once_huge;
-					 TestCase == server_echos_active_huge;
-					 TestCase == client_echos_passive_huge;
-					 TestCase == client_echos_active_once_huge;
-					 TestCase == client_echos_active_huge ->
+init_per_testcase(TestCase, Config)
+  when TestCase == server_echos_passive_huge;
+       TestCase == server_echos_passive_chunk_huge;
+       TestCase == server_echos_active_once_huge;
+       TestCase == server_echos_active_huge;
+       TestCase == client_echos_passive_huge;
+       TestCase == client_echos_passive_chunk_huge;
+       TestCase == client_echos_active_once_huge;
+       TestCase == client_echos_active_huge ->
     case erlang:system_info(system_architecture) of
 	"sparc-sun-solaris2.10" ->
 	    {skip,"Will take to long time on an old Sparc"};
 	_ ->
-	    ct:timetrap({seconds, 90}),
+	    ct:timetrap(?TIMEOUT_LONG),
 	    Config
     end;
 
-init_per_testcase(TestCase, Config) when TestCase == server_echos_passive_big;
-					 TestCase == server_echos_active_once_big;
-					 TestCase == server_echos_active_big;
-					 TestCase == client_echos_passive_big;
-					 TestCase == client_echos_active_once_big;
-					 TestCase == client_echos_active_big ->
-    ct:timetrap({seconds, 60}),
+init_per_testcase(TestCase, Config)
+  when TestCase == server_echos_passive_big;
+       TestCase == server_echos_passive_chunk_big;
+       TestCase == server_echos_active_once_big;
+       TestCase == server_echos_active_big;
+       TestCase == client_echos_passive_big;
+       TestCase == client_echos_passive_chunk_big;
+       TestCase == client_echos_active_once_big;
+       TestCase == client_echos_active_big ->
+    ct:timetrap(?TIMEOUT_LONG),
     Config;
 
 init_per_testcase(_TestCase, Config) ->
-    ct:timetrap({seconds, 15}),
+    ct:timetrap(?TIMEOUT),
     Config.
 
 end_per_testcase(_TestCase, Config) ->
@@ -144,11 +160,26 @@ server_echos_passive_small(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    server_echos_passive(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
+
+%%--------------------------------------------------------------------
+
+server_echos_passive_chunk_small() ->
+    [{doc, "Client sends 1000 bytes in passive mode to server, that receives them in chunks, "
+     "sends them back, and closes."}].
+
+server_echos_passive_chunk_small(Config) when is_list(Config) -> 
+    ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    server_echos_passive_chunk(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
-    Str = "1234567890", 
-    
-    server_echos_passive(Str, 1000, ClientOpts, ServerOpts, 
-			 ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 
@@ -160,11 +191,10 @@ server_echos_active_once_small(Config) when is_list(Config) ->
         ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    server_echos_active_once(Str, 1000, ClientOpts, ServerOpts, 
-			     ClientNode, ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    server_echos_active_once(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 
@@ -176,11 +206,10 @@ server_echos_active_small(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    server_echos_active(Str, 1000, ClientOpts, ServerOpts, 
-			ClientNode, ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    server_echos_active(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 client_echos_passive_small() ->
@@ -191,11 +220,25 @@ client_echos_passive_small(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    client_echos_passive(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
-    Str = "1234567890", 
+%%--------------------------------------------------------------------
+client_echos_passive_chunk__small() ->
+    [{doc, "Server sends 1000 bytes in passive mode to client, that receives them in chunks, "
+      "sends them back, and closes."}].
+
+client_echos_passive_chunk_small(Config) when is_list(Config) -> 
+    ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    client_echos_passive_chunk(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
-    client_echos_passive(Str, 1000, ClientOpts, ServerOpts, ClientNode, 
-			 ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 client_echos_active_once_small() ->
@@ -206,11 +249,10 @@ client_echos_active_once_small(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    client_echos_active_once(Str, 1000, ClientOpts, ServerOpts, ClientNode, 
-			     ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    client_echos_active_once(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
    
 %%--------------------------------------------------------------------
 client_echos_active_small() ->
@@ -221,11 +263,10 @@ client_echos_active_small(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890",     
-
-    client_echos_active(Str, 1000, ClientOpts, ServerOpts, ClientNode, 
-			ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 100),
+    client_echos_active(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 
 %%--------------------------------------------------------------------
@@ -237,11 +278,23 @@ server_echos_passive_big(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    server_echos_passive(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
+%%--------------------------------------------------------------------
+server_echos_passive_chunk_big() ->
+    [{doc, "Client sends 50000 bytes to server in passive mode, that receives them, "
+     "sends them back, and closes."}].
 
-    Str = "1234567890", 
-    
-    server_echos_passive(Str, 50000, ClientOpts, ServerOpts, ClientNode, 
-			 ServerNode, Hostname).
+server_echos_passive_chunk_big(Config) when is_list(Config) -> 
+    ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    server_echos_passive_chunk(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 
@@ -253,11 +306,10 @@ server_echos_active_once_big(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    server_echos_active_once(Str, 50000, ClientOpts, ServerOpts, ClientNode, 
-			     ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    server_echos_active_once(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 
@@ -269,11 +321,10 @@ server_echos_active_big(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    server_echos_active(Str, 50000, ClientOpts, ServerOpts, ClientNode, 
-			ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    server_echos_active(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 client_echos_passive_big() ->
@@ -284,11 +335,26 @@ client_echos_passive_big(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    client_echos_passive(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
-    Str = "1234567890", 
 
-    client_echos_passive(Str, 50000, ClientOpts, ServerOpts, ClientNode, 
-			 ServerNode, Hostname).
+%%--------------------------------------------------------------------
+client_echos_passive_chunk_big() ->
+    [{doc, "Server sends 50000 bytes to client in passive mode, that receives them, "
+     "sends them back, and closes."}].
+
+client_echos_passive_chunk_big(Config) when is_list(Config) -> 
+    ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    client_echos_passive_chunk(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
+
 
 %%--------------------------------------------------------------------
 client_echos_active_once_big() ->
@@ -299,11 +365,10 @@ client_echos_active_once_big(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-
-    client_echos_active_once(Str, 50000, ClientOpts, ServerOpts, ClientNode, 
-			     ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    client_echos_active_once(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
    
 %%--------------------------------------------------------------------
 client_echos_active_big() ->
@@ -314,11 +379,10 @@ client_echos_active_big(Config) when is_list(Config) ->
      ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    client_echos_active(Str, 50000, ClientOpts, ServerOpts, ClientNode, 
-			ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 5000),
+    client_echos_active(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 server_echos_passive_huge() ->
@@ -329,11 +393,24 @@ server_echos_passive_huge(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    server_echos_passive(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
-    Str = "1234567890", 
-    
-    server_echos_passive(Str, 500000, ClientOpts, ServerOpts, ClientNode, 
-			 ServerNode, Hostname).
+%%--------------------------------------------------------------------
+server_echos_passive_chunk_huge() ->
+    [{doc, "Client sends 500000 bytes to server in passive mode, that receives "
+      " them, sends them back, and closes."}].
+
+server_echos_passive_chunk_huge(Config) when is_list(Config) -> 
+    ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    server_echos_passive_chunk(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 server_echos_active_once_huge() ->
@@ -344,11 +421,10 @@ server_echos_active_once_huge(Config) when is_list(Config) ->
         ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    server_echos_active_once(Str, 500000, ClientOpts, ServerOpts, ClientNode, 
-			     ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    server_echos_active_once(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 server_echos_active_huge() ->
@@ -359,11 +435,10 @@ server_echos_active_huge(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    
-    server_echos_active(Str, 500000, ClientOpts, ServerOpts, ClientNode, 
-			ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    server_echos_active(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 
 %%--------------------------------------------------------------------
 client_echos_passive_huge() ->
@@ -374,11 +449,23 @@ client_echos_passive_huge(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    client_echos_passive(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
+%%--------------------------------------------------------------------
+client_echos_passive_chunk_huge() ->
+    [{doc, "Server sends 500000 bytes to client in passive mode, that receives "
+     "them, sends them back, and closes."}].
 
-    Str = "1234567890", 
-    client_echos_passive(Str, 500000, ClientOpts, ServerOpts, ClientNode, 
-			 ServerNode, Hostname).
-
+client_echos_passive_chunk_huge(Config) when is_list(Config) -> 
+    ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    client_echos_passive_chunk(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
 %%--------------------------------------------------------------------
 client_echos_active_once_huge() ->
     [{doc, "Server sends 500000 bytes to client in active once mode, that receives "
@@ -388,10 +475,10 @@ client_echos_active_once_huge(Config) when is_list(Config) ->
     ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    client_echos_active_once(Str, 500000, ClientOpts, ServerOpts, ClientNode, 
-			     ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    client_echos_active_once(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
    
 %%--------------------------------------------------------------------
 client_echos_active_huge() ->
@@ -402,293 +489,326 @@ client_echos_active_huge(Config) when is_list(Config) ->
      ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
     ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
-
-    Str = "1234567890", 
-    client_echos_active(Str, 500000, ClientOpts, ServerOpts, ClientNode,
-			ServerNode, Hostname).
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    client_echos_active(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
  
+
+%%--------------------------------------------------------------------
+client_active_once_server_close() ->
+    [{doc, "Server sends 500000 bytes and immediately after closes the connection"
+     "Make sure client recives all data if possible"}].
+
+client_active_once_server_close(Config) when is_list(Config) -> 
+    ClientOpts = ssl_test_lib:ssl_options(client_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(server_opts, Config),
+    {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
+    %%
+    Data = binary:copy(<<"1234567890">>, 50000),
+    client_active_once_server_close(
+      Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname).
+ 
+
+
 %%--------------------------------------------------------------------
 %% Internal functions ------------------------------------------------
 %%--------------------------------------------------------------------
 
-server_echos_passive(Data, Length, ClientOpts, ServerOpts, 
-		     ClientNode, ServerNode, Hostname) ->
-    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0}, 
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, echoer, 
-					  [Data, Length]}},
-					{options, 
-					 [{active, false},{mode, binary} 
-					  | ServerOpts]}]),
-				       Port = ssl_test_lib:inet_port(Server),
-    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port}, 
-					{host, Hostname},
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, sender, 
-					       [Data, 
-						Length]}},
-					{options, 
-					 [{active, false}, {mode, binary} |
-					  ClientOpts]}]),
-    ssl_test_lib:check_result(Server, ok, Client, ok),
-    
-    ssl_test_lib:close(Server),
-    ssl_test_lib:close(Client).
-
-
-server_echos_active_once(Data, Length, ClientOpts, ServerOpts, ClientNode,
-			 ServerNode, Hostname) ->
-    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0}, 
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, echoer_once, 
-					  [Data, Length]}},
-					{options, [{active, once},
-						   {mode, binary}| 
-						   ServerOpts]}]),
+server_echos_passive(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, echoer, [Length]}},
+           {options, [{active, false}, {mode, binary} | ServerOpts] ++ ssl_test_lib:bigger_buffers()}]),
     Port = ssl_test_lib:inet_port(Server),
-    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port}, 
-					{host, Hostname},
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, sender_once, 
-					  [Data, Length]}},
-					{options, [{active, once}, 
-						   {mode, binary} | 
-						   ClientOpts]}]),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, sender, [Data]}},
+           {options, [{active, false}, {mode, binary} | ClientOpts] ++ ssl_test_lib:bigger_buffers() }]),
+    %%
     ssl_test_lib:check_result(Server, ok, Client, ok),
-    
+    %%
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
 
-
-server_echos_active(Data, Length, ClientOpts, ServerOpts, 
-		    ClientNode, ServerNode, Hostname) ->
-    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0}, 
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, echoer_active, 
-					  [Data, Length]}},
-					{options,  
-					 [{active, true},
-					  {mode, binary} | ServerOpts]}]),
+server_echos_passive_chunk(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, echoer_chunk, [Length]}},
+           {options, [{active, false}, {mode, binary} | ServerOpts]}]),
     Port = ssl_test_lib:inet_port(Server),
-    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port}, 
-					{host, Hostname},
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, sender_active, 
-					       [Data, 
-						Length]}},
-					{options,  
-					 [{active, true}, {mode, binary}
-					  | ClientOpts]}]),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, sender, [Data]}},
+           {options, [{active, false}, {mode, binary} | ClientOpts]}]),
+    %%
     ssl_test_lib:check_result(Server, ok, Client, ok),
-    
+    %%
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
 
-client_echos_passive(Data, Length, ClientOpts, ServerOpts, 
-		     ClientNode, ServerNode, Hostname) ->
-    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0}, 
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, sender, 
-					  [Data, Length]}},
-					{options, 
-					 [{active, false}, {mode, binary} | 
-					  ServerOpts]}]),
+server_echos_active_once(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, echoer_active_once, [Length]}},
+           {options, [{active, once}, {mode, binary} | ServerOpts] ++ ssl_test_lib:bigger_buffers()}]),
     Port = ssl_test_lib:inet_port(Server),
-    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port}, 
-					{host, Hostname},
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, echoer, 
-					       [Data, 
-						Length]}},
-					{options, 
-					 [{active, false}, {mode, binary}
-					  | ClientOpts]}]),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, sender_active_once, [Data]}},
+           {options, [{active, once}, {mode, binary} | ClientOpts] ++ ssl_test_lib:bigger_buffers() }]),
+    %%
     ssl_test_lib:check_result(Server, ok, Client, ok),
-    
+    %%
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
 
-client_echos_active_once(Data, Length, 
-			 ClientOpts, ServerOpts, ClientNode, ServerNode,
-			 Hostname) ->
-    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0}, 
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, sender_once, 
-					  [Data, Length]}},
-					{options, [{active, once},
-						   {mode, binary} | 
-						   ServerOpts]}]),
+
+server_echos_active(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, echoer_active, [Length]}},
+           {options, [{active, true}, {mode, binary} | ServerOpts] ++ ssl_test_lib:bigger_buffers()}]),
     Port = ssl_test_lib:inet_port(Server),
-    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port}, 
-					{host, Hostname},
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, echoer_once, 
-					  [Data, 
-					   Length]}},
-					{options,[{active, once},
-						  {mode, binary}
-						  | ClientOpts]}]),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, sender_active, [Data]}},
+           {options, [{active, true}, {mode, binary} | ClientOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    %%
     ssl_test_lib:check_result(Server, ok, Client, ok),
-    
+    %%
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
 
-client_echos_active(Data, Length, ClientOpts, ServerOpts, ClientNode, 
-		    ServerNode,
-			 Hostname) ->
-    Server = ssl_test_lib:start_server([{node, ServerNode}, {port, 0}, 
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, sender_active, 
-					  [Data, Length]}},
-					{options, [{active, true}, 
-						   {mode, binary}
-						   | ServerOpts]}]),
+client_echos_passive(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, sender, [Data]}},
+           {options, [{active, false}, {mode, binary} | ServerOpts] ++ ssl_test_lib:bigger_buffers()}]),
     Port = ssl_test_lib:inet_port(Server),
-    Client = ssl_test_lib:start_client([{node, ClientNode}, {port, Port}, 
-					{host, Hostname},
-					{from, self()}, 
-					{mfa, 
-					 {?MODULE, echoer_active, 
-					       [Data, 
-						Length]}},
-					{options, [{active, true},
-						   {mode, binary}
-						   | ClientOpts]}]),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, echoer, [Length]}},
+           {options, [{active, false}, {mode, binary} | ClientOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    %%
     ssl_test_lib:check_result(Server, ok, Client, ok),
-    
+    %%
     ssl_test_lib:close(Server),
     ssl_test_lib:close(Client).
 
-send(_, _, _, 0,_) ->
+
+client_echos_passive_chunk(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, sender, [Data]}},
+           {options, [{active, false}, {mode, binary} | ServerOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    Port = ssl_test_lib:inet_port(Server),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, echoer_chunk, [Length]}},
+           {options, [{active, false}, {mode, binary} | ClientOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    %%
+    ssl_test_lib:check_result(Server, ok, Client, ok),
+    %%
+    ssl_test_lib:close(Server),
+    ssl_test_lib:close(Client).
+
+
+client_echos_active_once(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, sender_active_once, [Data]}},
+           {options, [{active, once}, {mode, binary} | ServerOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    Port = ssl_test_lib:inet_port(Server),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, echoer_active_once, [Length]}},
+           {options,[{active, once}, {mode, binary} | ClientOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    %%
+    ssl_test_lib:check_result(Server, ok, Client, ok),
+    %%
+    ssl_test_lib:close(Server),
+    ssl_test_lib:close(Client).
+
+client_echos_active(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, sender_active, [Data]}},
+           {options, [{active, true}, {mode, binary} | ServerOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    Port = ssl_test_lib:inet_port(Server),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {?MODULE, echoer_active, [Length]}},
+           {options, [{active, true}, {mode, binary} | ClientOpts] ++ ssl_test_lib:bigger_buffers()}]),
+    %
+    ssl_test_lib:check_result(Server, ok, Client, ok),
+    %%
+    ssl_test_lib:close(Server),
+    ssl_test_lib:close(Client).
+
+client_active_once_server_close(
+  Data, ClientOpts, ServerOpts, ClientNode, ServerNode, Hostname) ->
+    Length = byte_size(Data),
+    Server =
+        ssl_test_lib:start_server(
+          [{node, ServerNode}, {port, 0},
+           {from, self()},
+           {mfa, {?MODULE, send_close, [Data]}},
+           {options, [{active, once}, {mode, binary} | ServerOpts]}]),
+    Port = ssl_test_lib:inet_port(Server),
+    Client =
+        ssl_test_lib:start_client(
+          [{node, ClientNode}, {port, Port},
+           {host, Hostname},
+           {from, self()},
+           {mfa, {ssl_test_lib, active_once_recv, [Length]}},
+           {options,[{active, once}, {mode, binary} | ClientOpts]}]),
+    %%
+    ssl_test_lib:check_result(Server, ok, Client, ok).
+
+send(_Socket, _Data, 0, _) ->
     ok;
-send(Socket, Data, Size, Repeate,F) ->
-    NewData = lists:duplicate(Size div 10, Data),
-    ssl:send(Socket, NewData),
-    F(),
-    send(Socket, Data, Size, Repeate - 1,F).
- 
-sender(Socket, Data, Size) ->
-    ok = send(Socket, Data, Size, 100, fun() -> do_recv(Socket, Data, Size, <<>>, false) end),
+send(Socket, Data, Count, RecvEcho) ->
+    spawn(fun() -> ssl:send(Socket, Data) end),
+    RecvEcho(),
+    send(Socket, Data, Count - 1, RecvEcho).
+
+send_close(Socket, Data) ->
+    ok = ssl:send(Socket, Data),
+    ssl:close(Socket).
+
+sender(Socket, Data) ->
     ct:log("Sender recv: ~p~n", [ssl:getopts(Socket, [active])]),
-    ok.
+    send(Socket, Data, 100,
+              fun() -> 
+                      ssl_test_lib:recv_disregard(Socket, byte_size(Data)) 
+              end).
 
-sender_once(Socket, Data, Size) ->
-    send(Socket, Data, Size, 100, 
-	 fun() -> do_active_once(Socket, Data, Size, <<>>, false) end),
-    ct:log("Sender active once: ~p~n",
-		       [ssl:getopts(Socket, [active])]),
-    ok.
+sender_active_once(Socket, Data) ->
+    ct:log("Sender active once: ~p~n", [ssl:getopts(Socket, [active])]),
+    send(Socket, Data, 100,
+         fun() -> 
+                 ssl_test_lib:active_once_disregard(Socket, byte_size(Data)) 
+         end).
 
-sender_active(Socket, Data, Size) ->
-    F = fun() -> do_active(Socket, Data, Size, <<>>, false) end,
-    send(Socket, Data, Size, 100, F),
+sender_active(Socket, Data) ->
     ct:log("Sender active: ~p~n", [ssl:getopts(Socket, [active])]),
-    ok.
+    send(Socket, Data, 100,
+         fun() -> 
+                 ssl_test_lib:active_disregard(Socket, byte_size(Data)) 
+         end).
 
-echoer(Socket, Data, Size) ->
+echoer(Socket, Size) ->
     ct:log("Echoer recv: ~p~n", [ssl:getopts(Socket, [active])]),
-    echo(fun() -> do_recv(Socket, Data, Size, <<>>, true) end, 100).
+    echo_recv(Socket, Size * 100).
 
-echoer_once(Socket, Data, Size) ->
-    ct:log("Echoer active once: ~p ~n",
-		       [ssl:getopts(Socket, [active])]),
-    echo(fun() -> do_active_once(Socket, Data, Size, <<>>, true) end, 100).
+echoer_chunk(Socket, Size) ->
+    ct:log("Echoer recv: ~p~n", [ssl:getopts(Socket, [active])]),
+    echo_recv_chunk(Socket, Size, Size * 100).
 
-echoer_active(Socket, Data, Size) ->
+echoer_active_once(Socket, Size) ->
+    ct:log("Echoer active once: ~p~n", [ssl:getopts(Socket, [active])]),
+    echo_active_once(Socket, Size * 100).
+
+echoer_active(Socket, Size) ->
     ct:log("Echoer active: ~p~n", [ssl:getopts(Socket, [active])]),
-    echo(fun() -> do_active(Socket, Data, Size, <<>>, true) end, 100).
-
-echo(_Fun, 0) -> ok;
-echo(Fun, N) ->
-    Fun(),
-    echo(Fun, N-1).
+    echo_active(Socket, Size * 100).
 
 
-do_recv(_Socket, _Data, 0, _Acc, true) ->
+%% Receive Size bytes
+echo_recv(_Socket, 0) ->
     ok;
-do_recv(_Socket, Data, 0, Acc, false) ->
-    Data = lists:sublist(binary_to_list(Acc), 10);
+echo_recv(Socket, Size) ->
+    {ok, Data} = ssl:recv(Socket, 0),
+    spawn(fun() -> ssl:send(Socket, Data) end),
+    echo_recv(Socket, Size - byte_size(Data)).
 
-do_recv(Socket, Data, Size, Acc, Echo) ->
-    {ok, NewData} = ssl:recv(Socket, 0),
-    NewSize = size(NewData), 
-    case Echo of
-	true ->
-	    ssl:send(Socket, NewData),
-	    NewSize = size(NewData), 
-	    do_recv(Socket, Data, Size - NewSize, [], Echo);
-	false ->
-	    case size(Acc) < 10 of
-		true ->
-		    do_recv(Socket, Data, Size - NewSize, 
-			    <<Acc/binary, NewData/binary>>, Echo);
-		false ->
-		    do_recv(Socket, Data, Size - NewSize, Acc, Echo) 
-	    end
+
+%% Receive Size bytes
+echo_recv_chunk(_Socket, _, 0) ->
+    ok;
+echo_recv_chunk(Socket, ChunkSize, Size) ->
+    {ok, Data} = ssl:recv(Socket, ChunkSize),
+    spawn(fun() -> ssl:send(Socket, Data) end),
+    echo_recv_chunk(Socket, ChunkSize, Size - ChunkSize).
+
+
+%% Receive Size bytes
+echo_active_once(_Socket, 0) ->
+    ok;
+echo_active_once(Socket, Size) ->
+    receive
+        {ssl, Socket, Data} ->
+            spawn(fun() -> ssl:send(Socket, Data) end),
+            NewSize = Size - byte_size(Data),
+            ssl:setopts(Socket, [{active, once}]),
+            echo_active_once(Socket, NewSize)
     end.
 
-do_active_once(_Socket, _Data, 0, _Acc, true) ->
+%% Receive Size bytes
+echo_active(_Socket, 0) ->
     ok;
-do_active_once(_Socket, Data, 0, Acc, false) ->
-    Data = lists:sublist(binary_to_list(Acc), 10);
+echo_active(Socket, Size) ->
+    receive
+        {ssl, Socket, Data} ->
+            spawn(fun() -> ssl:send(Socket, Data) end),
+            echo_active(Socket, Size - byte_size(Data))
+    end.    
+        
 
-do_active_once(Socket, Data, Size, Acc, Echo) ->    
-    receive 
-	{ssl, Socket, NewData} ->
-	    NewSize = size(NewData), 
-	    case Echo of
-		true ->
-		    ssl:send(Socket, NewData),
-		    ssl:setopts(Socket, [{active, once}]),
-		    do_active_once(Socket, Data, Size - NewSize, [], Echo);
-		false ->
-		    case size(Acc) < 10 of
-			true ->
-			    ssl:setopts(Socket, [{active, once}]),
-			    do_active_once(Socket, Data, Size - NewSize, 
-					   <<Acc/binary, NewData/binary>>, 
-					   Echo);
-			false ->
-			    ssl:setopts(Socket, [{active, once}]),
-			    do_active_once(Socket, Data, 
-					   Size - NewSize, Acc, Echo) 
-		    end 
-	    end
-    end.
-    
-do_active(_Socket, _Data, 0, _Acc, true) ->
-    ok;
-do_active(_Socket, Data, 0, Acc, false) ->
-    Data = lists:sublist(binary_to_list(Acc), 10);
-
-do_active(Socket, Data, Size, Acc, Echo) ->    
-    receive 
-	{ssl, Socket, NewData} ->
-	    NewSize = size(NewData), 
-	    case Echo of
-		true ->
-		    ssl:send(Socket, NewData),
-		    do_active(Socket, Data, Size - NewSize, [], Echo);
-		false ->
-		    case size(Acc) < 10 of
-			true ->
-			    do_active(Socket, Data, Size - NewSize, 
-					   <<Acc/binary, NewData/binary>>, 
-				      Echo);
-			false ->
-			    do_active(Socket, Data, 
-				      Size - NewSize, Acc, Echo) 
-		    end 
-	    end
-    end.
